'''
instanceInfo.py provides information about an XBRL instance

See COPYRIGHT.md for copyright information.

Operation with arelleCmdLine: --plugin instanceInfo -f entryUrl

'''
import sys, os, time, math, logging
import regex as re
from math import isnan
from collections import defaultdict
from arelle.ValidateXbrlCalcs import inferredDecimals, rangeValue
from arelle import ModelDocument
from arelle.ModelInstanceObject import ModelFact
from arelle.Version import authorLabel, copyrightLabel
from arelle.XbrlConst import xhtml
from arelle.XmlUtil import ancestors, xmlstring

memoryAtStartup = 0
timeAtStart = 0
styleIxHiddenPattern = re.compile(r"(.*[^\w]|^)-(sec|esef)-ix-hidden\s*:\s*([\w.-]+).*")

def startup(cntlr, options, *args, **kwargs):
    global memoryAtStartup, timeAtStart
    memoryAtStartup = cntlr.memoryUsed
    timeAtStart = time.time()



def showInfo(cntlr, options, modelXbrl, _entrypoint, *args, **kwargs):
    for url, doc in sorted(modelXbrl.urlDocs.items(), key=lambda i: i[0]):
        if not any(url.startswith(w) for w in ("https://xbrl.sec.gov/", "http://xbrl.sec.gov/", "http://xbrl.fasb.org/", "http://www.xbrl.org/",
                                               "http://xbrl.ifrs.org/", "http://www.esma.europa.eu/")):
            if os.path.exists(doc.filepath): # skip if in an archive or stream
                cntlr.addToLog("File {} size {:,}".format(doc.basename, os.path.getsize(doc.filepath)), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Heap memory before loading {:,}".format(memoryAtStartup), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Heap memory after loading {:,}".format(cntlr.memoryUsed), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Time to load {:.2f} seconds".format(time.time() - timeAtStart), messageCode="info", level=logging.DEBUG)
    isInlineXbrl = modelXbrl.modelDocument.type in (ModelDocument.Type.INLINEXBRL, ModelDocument.Type.INLINEXBRLDOCUMENTSET)
    if isInlineXbrl:
        instanceType = "inline XBRL, number of documents {}".format(len(modelXbrl.ixdsHtmlElements))
    else:
        instanceType = "xBRL-XML"
    cntlr.addToLog("Instance type {}".format(instanceType), messageCode="info", level=logging.DEBUG)
    numContexts = len(modelXbrl.contexts)
    numLongContexts = 0
    bytesSaveableInline = 0
    bytesSaveableInlineWithCsv = 0
    frequencyOfDims = {}
    sumNumDims = 0
    distinctDurations = set()
    distinctInstants = set()
    shortContextIdLen = int(math.log10(numContexts or 1)) + 2 # if no contexts, use 1 for log function to work
    xbrlQnameCountInline = 0
    xbrlQnameCountInlineWithCsv = 0
    xbrlQnameLengthsInline = 0
    xbrlQnameLengthsInlineWithCsv = 0
    for c in modelXbrl.contexts.values():
        sumNumDims += len(c.qnameDims)
        for d in c.qnameDims.values():
            dimQname = str(d.dimensionQname)
            frequencyOfDims[dimQname] = frequencyOfDims.get(dimQname,0) + 1
            xbrlQnameCountInline += 1
            xbrlQnameCountInlineWithCsv += 1
            xbrlQnameLengthsInline += len(d.dimensionQname.localName)
            xbrlQnameLengthsInlineWithCsv += len(d.dimensionQname.localName)
        if c.isInstantPeriod:
            distinctInstants.add(c.instantDatetime)
        elif c.isStartEndPeriod:
            distinctDurations.add((c.startDatetime, c.endDatetime))
        if len(c.id) > shortContextIdLen:
            bytesSaveableInline += len(c.id) - shortContextIdLen
            bytesSaveableInlineWithCsv += len(c.id) - shortContextIdLen
    cntlr.addToLog("Number of contexts {:,}".format(numContexts), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of distinct durations {:,}".format(len(distinctDurations)), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of distinct instants {:,}".format(len(distinctInstants)), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Avg number dimensions per contexts {:,.2f}".format(sumNumDims/numContexts if numContexts else 0), messageCode="info", level=logging.DEBUG)
    mostPopularDims = sorted(frequencyOfDims.items(), key=lambda i:"{:0>9},{}".format(999999999-i[1],i[0]))
    for dimName, count in mostPopularDims[0:3]:
        cntlr.addToLog("Dimension {} used in {:,} contexts".format(dimName, count), messageCode="info", level=logging.DEBUG)

    # analyze for tables which could be composed from CSV data
    tblFacts = defaultdict(set)
    tblNestedTables = defaultdict(set)
    factSize = {}
    for f in modelXbrl.factsInInstance:
        for tdElt in ancestors(f, xhtml, "td"):
            factSize[f] = len(xmlstring(tdElt,stripXmlns=True))
            break
        childTblElt = None
        for tblElt in ancestors(f, xhtml, "table"):
            tblFacts[tblElt].add(f)
            if childTblElt:
                tblNestedTables[tblElt].add(childTblElt)

    # find tables containing only numeric facts
    def tblNestedFactCount(tbl):
        c = len(tblFacts.get(tbl, ()))
        for nestedTbl in tblNestedTables.get(tbl,()):
            c += tblNestedFactCount(nestedTbl)
        return c

    factsInInstance = len(modelXbrl.factsInInstance)
    factsInTables = len(set.union(*(fset for fset in tblFacts.values())))
    cntlr.addToLog("Facts in instance: {:,}, facts in tables: {:,}".format(factsInInstance,factsInTables), messageCode="info", level=logging.DEBUG)

    numTblsEligible = 0
    numFactsEligible = 0
    bytesCsvSavings = 0
    factsEligibleForCsv = set()
    tablesWithEligibleFacts = set()
    if tblFacts and factSize:
        # find eligible tables, have facts and not nested tables with other facts
        for tbl, facts in tblFacts.items():
            if len(facts) == tblNestedFactCount(tbl):
                s = sum(factSize.get(f,0) for f in facts) - sum(len(str(f.value)) for f in facts)
                if s > 10000:
                    numTblsEligible += 1
                    bytesCsvSavings += s
                    numFactsEligible += len(facts)
                    factsEligibleForCsv |= facts
                    tablesWithEligibleFacts.add(tbl)
    numFacts = 0
    numTableTextBlockFacts = 0
    lenTableTextBlockFacts = 0
    numTextBlockFacts = 0
    lenTextBlockFacts = 0
    distinctElementsInFacts = set()
    factsPerContext = {}
    factForConceptContextUnitHash = defaultdict(list)
    for f in modelXbrl.factsInInstance:
        context = f.context
        concept = f.concept
        distinctElementsInFacts.add(f.qname)
        numFacts += 1
        if f.qname.localName.endswith("TableTextBlock"):
            numTableTextBlockFacts += 1
            lenTableTextBlockFacts += len(f.xValue)
        elif f.qname.localName.endswith("TextBlock"):
            numTextBlockFacts += 1
            lenTextBlockFacts += len(f.xValue)
        if context is not None and concept is not None:
            factsPerContext[context.id] = factsPerContext.get(context.id,0) + 1
            factForConceptContextUnitHash[f.conceptContextUnitHash].append(f)
            bytesSaveableInline += len(context.id) - shortContextIdLen
            if f not in factsEligibleForCsv:
                bytesSaveableInlineWithCsv += len(context.id) - shortContextIdLen


    if numTblsEligible:
        cntlr.addToLog("Tables eligible for facts in CSV: {:,}, facts eligible for CSV: {:,}, bytes saveable by facts in CSV {:,}".format(numTblsEligible, numFactsEligible, bytesCsvSavings), messageCode="info", level=logging.DEBUG)
    else:
        cntlr.addToLog("No tables eligible for facts in CSV", messageCode="info", level=logging.DEBUG)


    mostPopularContexts = sorted(factsPerContext.items(), key=lambda i:"{:0>9},{}".format(999999999-i[1],i[0]))
    cntlr.addToLog("Number of facts {:,}".format(numFacts), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of TableTextBlock facts {:,} avg len {:,.0f}".format(numTableTextBlockFacts, lenTableTextBlockFacts/numTableTextBlockFacts if numTableTextBlockFacts else 0), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of TextBlock facts {:,} avg len {:,.0f}".format(numTextBlockFacts, lenTextBlockFacts/numTableTextBlockFacts if numTableTextBlockFacts else 0), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Max number facts per context {:,}".format(mostPopularContexts[0][1] if mostPopularContexts else 0), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Avg number facts per context {:,.2f}".format(sum([v for v in factsPerContext.values()])/numContexts if numContexts else 0), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Distinct elements in facts {:,}".format(len(distinctElementsInFacts)), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of bytes saveable context id of {} length is {:,}".format(shortContextIdLen, bytesSaveableInline), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Excepting facts eligible for CSV, number of bytes saveable context id of {} length is {:,}".format(shortContextIdLen, bytesSaveableInlineWithCsv), messageCode="info", level=logging.DEBUG)

    aspectEqualFacts = defaultdict(list)
    decVals = {}
    numConsistentDupFacts = numInConsistentDupFacts = 0
    for hashEquivalentFacts in factForConceptContextUnitHash.values():
        if len(hashEquivalentFacts) > 1:
            for f in hashEquivalentFacts:
                aspectEqualFacts[(f.qname,f.contextID,f.unitID,
                                  f.xmlLang.lower() if f.concept.type.isWgnStringFactType else None)].append(f)
            for fList in aspectEqualFacts.values():
                f0 = fList[0]
                if f0.concept.isNumeric:
                    if any(f.isNil for f in fList):
                        _inConsistent = not all(f.isNil for f in fList)
                    else: # not all have same decimals
                        _d = inferredDecimals(f0)
                        _v = f0.xValue
                        _inConsistent = isnan(_v) # NaN is incomparable, always makes dups inconsistent
                        decVals[_d] = _v
                        aMax, bMin, _inclA, _inclB = rangeValue(_v, _d)
                        for f in fList[1:]:
                            _d = inferredDecimals(f)
                            _v = f.xValue
                            if isnan(_v):
                                _inConsistent = True
                                break
                            if _d in decVals:
                                _inConsistent |= _v != decVals[_d]
                            else:
                                decVals[_d] = _v
                            a, b, _inclA, _inclB = rangeValue(_v, _d)
                            if a > aMax: aMax = a
                            if b < bMin: bMin = b
                        if not _inConsistent:
                            _inConsistent = (bMin < aMax)
                        decVals.clear()
                else:
                    _inConsistent = any(not f.isVEqualTo(f0) for f in fList[1:])
                if _inConsistent:
                    numInConsistentDupFacts += 1
                else:
                    numConsistentDupFacts += 1

            aspectEqualFacts.clear()
    cntlr.addToLog("Number of duplicate facts consistent {:,} inconsistent {:,}".format(numConsistentDupFacts, numInConsistentDupFacts), messageCode="info", level=logging.DEBUG)

    styleAttrCountsInline = {}
    styleAttrCountsInlineWithCsv = {}
    totalStyleLenInline = 0
    totalStyleLenInlineWithCsv = 0
    continuationElements = {}
    ixNsPrefix = "{http://www.xbrl.org/2013/inlineXBRL}"
    for ixdsHtmlRootElt in getattr(modelXbrl, "ixdsHtmlElements", ()): # ix root elements if inline
        for ixElt in ixdsHtmlRootElt.iterdescendants():
            inEligibleTableForCsv = any(p in tablesWithEligibleFacts for p in ixElt.iterancestors("{http://www.w3.org/1999/xhtml}table"))
            style = ixElt.get("style")
            ixEltTag = str(ixElt.tag)
            if style:
                styleAttrCountsInline[style] = styleAttrCountsInline.get(style,0) + 1
                if not inEligibleTableForCsv:
                    styleAttrCountsInlineWithCsv[style] = styleAttrCountsInlineWithCsv.get(style,0) + 1
                if styleIxHiddenPattern.match(style) is None:
                    totalStyleLenInline += len(style)
                    if not inEligibleTableForCsv:
                        totalStyleLenInlineWithCsv += len(style)
            if ixEltTag == "{http://www.xbrl.org/2013/inlineXBRL}continuation" and ixElt.id:
                continuationElements[ixElt.id] = ixElt
            if ixEltTag.startswith(ixNsPrefix):
                localName = ixEltTag[len(ixNsPrefix):]
                if localName == "continuation" and ixElt.id:
                    continuationElements[ixElt.id] = ixElt
                elif localName in ("nonFraction", "nonNumeric", "fraction"):
                    xbrlQnameCountInline += 1
                    xbrlQnameLengthsInline += len(ixElt.qname.localName)
                    if not inEligibleTableForCsv:
                        xbrlQnameCountInlineWithCsv += 1
                        xbrlQnameLengthsInlineWithCsv += len(ixElt.qname.localName)
            elif isinstance(ixElt, ModelFact):
                xbrlQnameCountInline += 2
                xbrlQnameLengthsInline += len(ixElt.qname.localName)
                if not inEligibleTableForCsv:
                    xbrlQnameCountInlineWithCsv += 2
                    xbrlQnameLengthsInlineWithCsv += len(ixElt.qname.localName)

    def locateContinuation(element, chain=None):
        contAt = element.get("continuedAt")
        if contAt:
            if contAt in continuationElements:
                if chain is None: chain = [element]
                contElt = continuationElements[contAt]
                if contElt not in chain:
                    chain.append(contElt)
                    element._continuationElement = contElt
                    return locateContinuation(contElt, chain)
        elif chain: # end of chain
            return len(chain)

    numContinuations = 0
    maxLenLen = 0
    maxLenHops = 0
    maxHops = 0
    maxHopsLen = 0
    for f in modelXbrl.factsInInstance:
        if f.get("continuedAt"):
            numContinuations += 1
            _len = len(f.xValue)
            _hops = locateContinuation(f)
            if _hops > maxHops:
                maxHops = _hops
                maxHopsLen = _len
            if _len > maxLenLen:
                maxLenLen = _len
                maxLenHops = _hops

    cntlr.addToLog("Number of continuation facts {:,}".format(numContinuations), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Longest continuation fact {:,} number of hops {:,}".format(maxLenLen, maxLenHops), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Most continuation hops {:,} fact len {:,}".format(maxHops, maxHopsLen), messageCode="info", level=logging.DEBUG)

    numDupStyles = sum(1 for n in styleAttrCountsInline.values() if n > 1)
    bytesSaveableByCssInline = sum(len(s)*(n-1) for s,n in styleAttrCountsInline.items() if n > 1)
    cntlr.addToLog("Number of duplicate styles {:,}, bytes saveable by CSS {:,}, len of all non-ix-hidden @styles {:,}".format(numDupStyles, bytesSaveableByCssInline, totalStyleLenInline), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Number of XBRL QNames {:,}, bytes saveable by EBA-style element names {:,}".format(xbrlQnameCountInline, xbrlQnameLengthsInline - (5*xbrlQnameCountInline)), messageCode="info", level=logging.DEBUG)
    numDupStyles = sum(1 for n in styleAttrCountsInlineWithCsv.values() if n > 1)
    bytesSaveableByCssInlineWithCsv = sum(len(s)*(n-1) for s,n in styleAttrCountsInlineWithCsv.items() if n > 1)
    cntlr.addToLog("Excepting facts eligible for CSV, number of duplicate styles {:,}, bytes saveable by CSS {:,}, len of all non-ix-hidden @styles {:,}".format(numDupStyles, bytesSaveableByCssInlineWithCsv, totalStyleLenInlineWithCsv), messageCode="info", level=logging.DEBUG)
    cntlr.addToLog("Excepting facts eligible for CSV, number of XBRL QNames {:,}, bytes saveable by EBA-style element names {:,}".format(xbrlQnameCountInlineWithCsv, xbrlQnameLengthsInlineWithCsv - (5*xbrlQnameCountInlineWithCsv)), messageCode="info", level=logging.DEBUG)



__pluginInfo__ = {
    'name': 'Instance Info',
    'version': '1.0',
    'description': "This plug-in displays instance information for sizing and performance issues.",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    'import': ('inlineXbrlDocumentSet',),
    # classes of mount points (required)
    'CntlrCmdLine.Filing.Start': startup,
    'CntlrCmdLine.Xbrl.Loaded': showInfo
}



'''
xbrlDB is an interface to XBRL databases.

Two implementations are provided:

(1) the XBRL Public Database schema for Postgres, published by XBRL US.

(2) an graph database, based on the XBRL Abstract Model PWD 2.

See COPYRIGHT.md for copyright information.
(Does not apply to the XBRL US Database schema and description.)

'''

import time, os, io, sys, logging
from arelle.Locale import format_string
from arelle.Version import authorLabel, copyrightLabel
from .XbrlPublicPostgresDB import insertIntoDB as insertIntoPostgresDB, isDBPort as isPostgresPort
from .XbrlSemanticSqlDB import insertIntoDB as insertIntoSemanticSqlDB, isDBPort as isSemanticSqlPort
from .XbrlOpenSqlDB import insertIntoDB as insertIntoOpenSqlDB
from .XbrlSemanticGraphDB import insertIntoDB as insertIntoRexsterDB, isDBPort as isRexsterPort
from .XbrlSemanticRdfDB import insertIntoDB as insertIntoRdfDB, isDBPort as isRdfPort
from .XbrlSemanticJsonDB import insertIntoDB as insertIntoJsonDB, isDBPort as isJsonPort
from .XbrlDpmSqlDB import insertIntoDB as insertIntoDpmDB, isDBPort as isDpmPort

dbTypes = {
    "postgres": insertIntoPostgresDB,
    "mssqlSemantic": insertIntoSemanticSqlDB,
    "mysqlSemantic": insertIntoSemanticSqlDB,
    "orclSemantic": insertIntoSemanticSqlDB,
    "pgSemantic": insertIntoSemanticSqlDB,
    "sqliteSemantic": insertIntoSemanticSqlDB,
    "sqliteDpmDB": insertIntoDpmDB,
    "pgOpenDB": insertIntoOpenSqlDB,
    "rexster": insertIntoRexsterDB,
    "rdfDB": insertIntoRdfDB,
    "json": insertIntoJsonDB
    }

dbProduct = {
    "postgres": "postgres",
    "mssqlSemantic": "mssql",
    "mysqlSemantic": "mysql",
    "orclSemantic": "orcl",
    "pgSemantic": "postgres",
    "pgOpenDB": "postgres",
    "sqliteSemantic": "sqlite",
    "sqliteDpmDB": "sqlite",
    "rexster": None,
    "rdfDB": None,
    "json": None
    }

_loadFromDBoptions = None  # only set for load, vs store operation
_storeIntoDBoptions = None
_schemaRefSubstitutions = None # for DPM database

def xbrlDBmenuEntender(cntlr, menu, *args, **kwargs):

    def storeIntoDBMenuCommand():
        # save DTS menu item has been invoked
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No XBRL instance or taxonomy is loaded.")
            return
        from arelle.DialogUserPassword import askDatabase
        # (user, password, host, port, database)
        priorDBconnection = cntlr.config.get("xbrlDBconnection", None)
        dbConnection = askDatabase(cntlr.parent, priorDBconnection)
        if not dbConnection: # action cancelled
            return

        def backgroundStoreIntoDB():
            try:
                host, port, user, password, db, timeout, dbType = dbConnection
                product = None
                if timeout and timeout.isdigit():
                    timeout = int(timeout)
                # identify server
                if dbType in dbTypes:
                    insertIntoDB = dbTypes[dbType]
                    product = dbProduct[dbType]
                else:
                    cntlr.addToLog(_("Probing host {0} port {1} to determine server database type.")
                                   .format(host, port))
                    if isPostgresPort(host, port):
                        dbType = "postgres"
                        insertIntoDB = insertIntoPostgresDB
                    elif isSemanticSqlPort(host, port):
                        dbType = "pgSemantic"
                        insertIntoDB = insertIntoPostgresDB
                    elif isRexsterPort(host, port):
                        dbType = "rexster"
                        insertIntoDB = insertIntoRexsterDB
                    elif isRdfPort(host, port, db):
                        dbType = "rdfDB"
                        insertIntoDB = insertIntoRdfDB
                    elif isJsonPort(host, port, db):
                        dbType = "json"
                        insertIntoDB = insertIntoJsonDB
                    else:
                        cntlr.addToLog(_("Unable to determine server type!\n  ") +
                                       _("Probing host {0} port {1} unable to determine server type.")
                                               .format(host, port))
                        cntlr.config["xbrlDBconnection"] = (host, port, user, password, db, timeout, '') # forget type
                        cntlr.saveConfig()
                        return
                    cntlr.addToLog(_("Database type {} identified.").format(dbType))
                cntlr.config["xbrlDBconnection"] = (host, port, user, password, db, timeout, dbType)
                cntlr.saveConfig()
                startedAt = time.time()
                insertIntoDB(cntlr.modelManager.modelXbrl,
                             host=host, port=port, user=user, password=password, database=db, timeout=timeout,
                             product=product)
                cntlr.addToLog(format_string(cntlr.modelManager.locale,
                                            _("stored to database in %.2f secs"),
                                            time.time() - startedAt))
            except Exception as ex:
                import traceback
                cntlr.addToLog(
                    _("[xpDB:exception] Loading XBRL DB: %(exception)s: %(error)s \n%(traceback)s") %
                    {"exception": ex.__class__.__name__,
                     "error": str(ex),
                     "exc_info": True,
                     "traceback": traceback.format_tb(sys.exc_info()[2])})
                cntlr.config["xbrlDBconnection"] = (host, port, user, password, db, timeout, '') # forget type
                cntlr.saveConfig()
        import threading
        thread = threading.Thread(target=backgroundStoreIntoDB)
        thread.daemon = True
        thread.start()

    # Extend menu with an item for the savedts plugin
    menu.add_command(label="Store to XBRL DB",
                     underline=0,
                     command=storeIntoDBMenuCommand)

    # add log handler
    logging.getLogger("arelle").addHandler(LogToDbHandler())

def storeIntoDB(dbConnection, modelXbrl, rssItem=None, **kwargs):
    host = port = user = password = db = timeout = dbType = None
    if isinstance(dbConnection, (list, tuple)): # variable length list
        if len(dbConnection) > 0: host = dbConnection[0]
        if len(dbConnection) > 1: port = dbConnection[1]
        if len(dbConnection) > 2: user = dbConnection[2]
        if len(dbConnection) > 3: password = dbConnection[3]
        if len(dbConnection) > 4: db = dbConnection[4]
        if len(dbConnection) > 5 and dbConnection[5] and dbConnection[5].isdigit():
            timeout = int(dbConnection[5])
        if len(dbConnection) > 6: dbType = dbConnection[6]

    startedAt = time.time()
    product = None
    if dbType in dbTypes:
        insertIntoDB = dbTypes[dbType]
        product = dbProduct[dbType]
    elif isPostgresPort(host, port):
        insertIntoDB = insertIntoPostgresDB
    elif isSemanticSqlPort(host, port):
        insertIntoDB = insertIntoSemanticSqlDB
    elif isRexsterPort(host, port):
        insertIntoDB = insertIntoRexsterDB
    elif isRdfPort(host, port, db):
        insertIntoDB = insertIntoRdfDB
    elif isJsonPort(host, port, db):
        insertIntoDB = insertIntoJsonDB
    else:
        modelXbrl.modelManager.addToLog('Server at "{0}:{1}" is not recognized to be either a Postgres or a Rexter service.'.format(host, port))
        return
    result = insertIntoDB(modelXbrl, host=host, port=port, user=user, password=password, database=db, timeout=timeout, product=product, rssItem=rssItem, **kwargs)
    if kwargs.get("logStoredMsg", result): # if false/None result and no logStoredMsg parameter then skip the message
        modelXbrl.modelManager.addToLog(format_string(modelXbrl.modelManager.locale,
                              _("stored to database in %.2f secs"),
                              time.time() - startedAt), messageCode="info", file=modelXbrl.uri)
    return result

def xbrlDBcommandLineOptionExtender(parser, *args, **kwargs):
    # extend command line options to store to database
    parser.add_option("--store-to-XBRL-DB",
                      action="store",
                      dest="storeIntoXbrlDb",
                      help=_("Store into XBRL DB.  "
                             "Provides connection string: host,port,user,password,database[,timeout[,{postgres|rexster|rdfDB}]]. "
                             "Autodetects database type unless 7th parameter is provided.  "))
    parser.add_option("--load-from-XBRL-DB",
                      action="store",
                      dest="loadFromXbrlDb",
                      help=_("Load from XBRL DB.  "
                             "Provides connection string: host,port,user,password,database[,timeout[,{postgres|rexster|rdfDB}]]. "
                             "Specifies DB parameters to load and optional file to save XBRL into.  "))

    logging.getLogger("arelle").addHandler(LogToDbHandler())

def xbrlDBCommandLineXbrlLoaded(cntlr, options, modelXbrl, entrypoint, *args, **kwargs):
    from arelle.ModelDocument import Type
    if modelXbrl.modelDocument.type == Type.RSSFEED and getattr(options, "storeIntoXbrlDb", False):
        modelXbrl.xbrlDBconnection = options.storeIntoXbrlDb.split(",")
        # for semantic SQL database check for loaded filings
        if (len(modelXbrl.xbrlDBconnection) > 7 and
            modelXbrl.xbrlDBconnection[6] in ("mssqlSemantic","mysqlSemantic","orclSemantic",
                                              "pgSemantic","sqliteSemantic","pgOpenDB") and
            modelXbrl.xbrlDBconnection[7] == "skipLoadedFilings"):
            # specify reloading of cached source documents (may have been corrupted originally or refiled)
            modelXbrl.reloadCache = True
            storeIntoDB(modelXbrl.xbrlDBconnection, modelXbrl, entrypoint=entrypoint, rssObject=modelXbrl.modelDocument)

def xbrlDBCommandLineXbrlRun(cntlr, options, modelXbrl, entrypoint, *args, **kwargs):
    from arelle.ModelDocument import Type
    if (modelXbrl.modelDocument.type not in (Type.RSSFEED, Type.TESTCASE, Type.REGISTRYTESTCASE) and
        getattr(options, "storeIntoXbrlDb", False) and
        not getattr(modelXbrl, "xbrlDBprocessedByStreaming", False)):
        dbConnection = options.storeIntoXbrlDb.split(",")
        storeIntoDB(dbConnection, modelXbrl, entrypoint=entrypoint)

def xbrlDBvalidateRssItem(val, modelXbrl, rssItem, *args, **kwargs):
    if hasattr(val.modelXbrl, 'xbrlDBconnection'):
        storeIntoDB(val.modelXbrl.xbrlDBconnection, modelXbrl, rssItem)

def xbrlDBtestcaseVariationXbrlLoaded(val, modelXbrl, *args, **kwargs):
    if _storeIntoDBoptions:
        return storeIntoDB(_storeIntoDBoptions.split(','), modelXbrl)

def xbrlDBdialogRssWatchDBconnection(*args, **kwargs):
    try:
        from .DialogRssWatchExtender import dialogRssWatchDBextender
        dialogRssWatchDBextender(*args, **kwargs)
    except ImportError:
        pass

def xbrlDBdialogRssWatchValidateChoices(dialog, frame, row, *args, **kwargs):
    from arelle.UiUtil import checkbox
    dialog.checkboxes += (
       checkbox(frame, 2, row,
                "Store into XBRL Database",
                "storeIntoXbrlDb"),
    )

def xbrlDBrssWatchHasWatchAction(rssWatchOptions, *args, **kwargs):
    return rssWatchOptions.get("xbrlDBconnection") and rssWatchOptions.get("storeIntoXbrlDB")

def xbrlDBrssDoWatchAction(modelXbrl, rssWatchOptions, rssItem, *args, **kwargs):
    dbConnectionString = rssWatchOptions.get("xbrlDBconnection")
    if dbConnectionString:
        dbConnection = dbConnectionString.split(',')
        storeIntoDB(dbConnection, modelXbrl)

def xbrlDBLoaderSetup(cntlr, options, *args, **kwargs):
    global _loadFromDBoptions, _storeIntoDBoptions, _schemaRefSubstitutions
    # set options to load from DB (instead of load from XBRL and store in DB)
    _loadFromDBoptions = getattr(options, "loadFromXbrlDb", None)
    _storeIntoDBoptions = getattr(options, "storeIntoXbrlDb", None)
    _schemaRefSubstitutions = None
    if _storeIntoDBoptions:
        dbConnection = _storeIntoDBoptions.split(',')
        if len(dbConnection) > 7 and dbConnection[6] == "sqliteDpmDB":
            for extraArg in dbConnection[7:]:
                argName, _sep, argValue = extraArg.partition("=")
                if argName == "schemaRefSubstitutions":
                    _schemaRefSubstitutions = dict(_keyVal.split(":")[0:2] for _keyVal in argValue.split(";"))

def xbrlDBLoader(modelXbrl, mappedUri, filepath, *args, **kwargs):
    # check if big instance and has header with an initial incomplete tree walk (just 2 elements
    if not _loadFromDBoptions:
        return None

    # load from DB and save XBRL in filepath, returning modelDocument
    extraArgs = {"loadDBsaveToFile": filepath, "logStoredMsg": False}
    dbConnection = _loadFromDBoptions.split(',')
    if len(dbConnection) > 7:
        for extraArg in dbConnection[7:]:
            argName, _sep, argValue = extraArg.partition("=")
            extraArgs[argName] = argValue
    return storeIntoDB(dbConnection, modelXbrl, **extraArgs)

def xbrlDBmodelXbrlInit(modelXbrl, *args, **kwargs):
        modelXbrl.xbrlDBprocessedByStreaming = False

def xbrlDBstartStreaming(modelXbrl, *args, **kwargs):
    if _storeIntoDBoptions:
        return storeIntoDB(_storeIntoDBoptions.split(','), modelXbrl, streamingState="start", logStoredMsg=False)

def xbrlDBstreamingFacts(modelXbrl, modelFacts, *args, **kwargs):
    if _storeIntoDBoptions:
        return storeIntoDB(_storeIntoDBoptions.split(','), modelXbrl, streamingState="acceptFacts", streamedFacts=modelFacts, logStoredMsg=False)

def xbrlDBfinishStreaming(modelXbrl, *args, **kwargs):
    if _storeIntoDBoptions:
        modelXbrl.xbrlDBprocessedByStreaming = True
        return storeIntoDB(_storeIntoDBoptions.split(','), modelXbrl, streamingState="finish", logStoredMsg=False)

def modelDocumentInstanceSchemaRefRewriter(modelDocument, url, *args, **kwargs):
    if _schemaRefSubstitutions:
        for _from, _to in _schemaRefSubstitutions.items():
            url = url.replace(_from, _to) # for DPM db substitutions
    return url

class LogToDbHandler(logging.Handler):
    def __init__(self):
        super(LogToDbHandler, self).__init__()
        self.logRecordBuffer = []

    def flush(self):
        del self.logRecordBuffer[:]

    def dbHandlerLogEntries(self, clear=True):
        entries = []
        for logRec in self.logRecordBuffer:
            message = { "text": self.format(logRec) }
            if logRec.args:
                for n, v in logRec.args.items():
                    message[n] = v
            entry = {"code": logRec.messageCode,
                     "level": logRec.levelname.lower(),
                     "refs": logRec.refs,
                     "message": message}
            entries.append(entry)
        if clear:
            del self.logRecordBuffer[:]
        return entries

    def emit(self, logRecord):
        self.logRecordBuffer.append(logRecord)


__pluginInfo__ = {
    'name': 'xbrlDB',
    'version': '0.9',
    'description': "This plug-in implements the XBRL Public Postgres, Abstract Model and DPM Databases.  ",
    'aliases': [
        'XBRL Database',
    ],
    'license': 'Apache-2 (Arelle plug-in), BSD license (pg8000 library)',
    'author': authorLabel,
    'copyright': f'{copyrightLabel} \n'
                'uses: cx_Oracle Copyright (c) 2007-2012, Anthony Tuininga. All rights reserved (Oracle DB), \n'
                '           (and)Copyright (c) 2001-2007, Computronix (Canada) Ltd., Edmonton, Alberta, Canada. All rights reserved, \n'
                '      pg8000, Copyright (c) 2007-2009, Mathieu Fenniak (Postgres DB), \n'
                '      pyodbc, no copyright, Michael Kleehammer (MS SQL), \n'
                '      PyMySQL, Copyright (c) 2010, 2013 PyMySQL contributors (MySQL DB), and\n'
                '      rdflib, Copyright (c) 2002-2012, RDFLib Team (RDF DB)',
    # classes of mount points (required)
    'CntlrWinMain.Menu.Tools': xbrlDBmenuEntender,
    'CntlrCmdLine.Options': xbrlDBcommandLineOptionExtender,
    'CntlrCmdLine.Utility.Run': xbrlDBLoaderSetup,
    'CntlrCmdLine.Xbrl.Loaded': xbrlDBCommandLineXbrlLoaded,
    'CntlrCmdLine.Xbrl.Run': xbrlDBCommandLineXbrlRun,
    'DialogRssWatch.FileChoices': xbrlDBdialogRssWatchDBconnection,
    'DialogRssWatch.ValidateChoices': xbrlDBdialogRssWatchValidateChoices,
    'ModelDocument.PullLoader': xbrlDBLoader,
    'RssWatch.HasWatchAction': xbrlDBrssWatchHasWatchAction,
    'RssWatch.DoWatchAction': xbrlDBrssDoWatchAction,
    'ModelXbrl.Init': xbrlDBmodelXbrlInit,
    'Streaming.Start': xbrlDBstartStreaming,
    'Streaming.Facts': xbrlDBstreamingFacts,
    'Streaming.Finish': xbrlDBfinishStreaming,
    'Validate.RssItem': xbrlDBvalidateRssItem,
    'TestcaseVariation.Xbrl.Loaded': xbrlDBtestcaseVariationXbrlLoaded,
    'ModelDocument.InstanceSchemaRefRewriter': modelDocumentInstanceSchemaRefRewriter
}



from pathlib import PurePath, Path
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig, ConformanceSuiteAssetConfig

config = ConformanceSuiteConfig(
    args=[
        '--disclosureSystem', 'hmrc',
    ],
    assets=[
        ConformanceSuiteAssetConfig.local_conformance_suite(
            Path('HMRC'),
            entry_point=Path('index.xml'),
        ),
    ] + [
        # https://www.frc.org.uk/library/standards-codes-policy/accounting-and-reporting/frc-taxonomies/historical-frc-taxonomy-suites/2021-frc-taxonomy-suite/
        ConformanceSuiteAssetConfig.public_taxonomy_package(
            Path('FRC_2021_Taxonomy_v1-0-0_Zip_File.zip'),
            public_download_url='https://www.frc.org.uk/documents/3923/FRC_2021_Taxonomy_v1-0-0_Zip_File.zip'
        ),
        # https://www.frc.org.uk/library/standards-codes-policy/accounting-and-reporting/frc-taxonomies/current-frc-taxonomy-suites/2023-frc-taxonomy-suite/
        ConformanceSuiteAssetConfig.public_taxonomy_package(
            Path('The_2023_Taxonomy_suite_v1.0.1.zip'),
            public_download_url='https://www.frc.org.uk/documents/372/The_2023_Taxonomy_suite_v1.0.1.zip'
        ),
        # https://www.frc.org.uk/library/standards-codes-policy/accounting-and-reporting/frc-taxonomies/current-frc-taxonomy-suites/2023-frc-taxonomy-suite/
        ConformanceSuiteAssetConfig.public_taxonomy_package(
            Path('Charities_2023_Taxonomies_2023.zip'),
            public_download_url='https://www.frc.org.uk/documents/3421/Charities_2023_Taxonomies_2023.zip'
        ),
    ],
    info_url='https://www.gov.uk/government/organisations/hm-revenue-customs',
    name=PurePath(__file__).stem,
    network_or_cache_required=False,
    plugins=frozenset({'validate/UK'}),
    shards=4,
    test_case_result_options='match-any',
)



'''
See COPYRIGHT.md for copyright information.
'''
import os, io, logging
from collections import defaultdict
from typing import Optional

from arelle import XmlUtil, XbrlConst, ModelValue
from arelle.ModelObject import ModelObject
from arelle.PluginManager import pluginClassMethods

TXMY_PKG_SRC_ELTS = ("metadata", "catalog", "taxonomy")


def testcaseVariationsByTarget(testcaseVariations):
    for modelTestcaseVariation in testcaseVariations:
        modelTestcaseVariation.errors = None # Errors accumulate over multiple ixdsTargets for same variation
        ixdsTargets = [instElt.get("target")
                      for resultElt in modelTestcaseVariation.iterdescendants("{*}result")
                      for instElt in resultElt.iterdescendants("{*}instance")]
        if ixdsTargets:
            # track status and actual (error codes, counts) across all targets
            allTargetsActual = []
            allTargetsStatus = ""
            for ixdsTarget in ixdsTargets:
                modelTestcaseVariation.ixdsTarget = ixdsTarget
                yield modelTestcaseVariation
                allTargetsActual.extend(modelTestcaseVariation.actual)
                if allTargetsStatus not in ("fail", "fail (count)"):
                    # update status unless fail were noted by a prior target of this variation
                    allTargetsStatus = modelTestcaseVariation.status
            modelTestcaseVariation.status = allTargetsStatus
        else: # probably an expected error situation
            modelTestcaseVariation.ixdsTarget = None
            yield modelTestcaseVariation

class ModelTestcaseVariation(ModelObject):
    def init(self, modelDocument):
        super(ModelTestcaseVariation, self).init(modelDocument)
        self.status = ""
        self.duration: float | None = None
        self.actual = []
        self.assertions = None
        self.ixdsTarget = None

    @property
    def id(self):
        # if there is a real ID, use it
        id = super(ModelTestcaseVariation, self).id
        if id is not None:
            return id
        # no ID, use the object ID so it isn't None
        return self.objectId()

    @property
    def name(self):
        try:
            return self._name
        except AttributeError:
            if self.get("name"):
                self._name = self.get("name")
            else:
                nameElement = XmlUtil.descendant(self, None, "name" if self.localName != "testcase" else "number")
                if nameElement is not None:
                    self._name = XmlUtil.innerText(nameElement)
                else:
                    self._name = None
            return self._name

    @property
    def description(self):
        nameElement = XmlUtil.descendant(self, None, ("description", "documentation"))
        if nameElement is not None:
            return XmlUtil.innerText(nameElement)
        return None

    @property
    def reference(self):
        efmNameElts = XmlUtil.children(self.getparent(), None, "name")
        for efmNameElt in efmNameElts:
            if efmNameElt is not None and efmNameElt.text.startswith("EDGAR"):
                return efmNameElt.text
        referenceElement = XmlUtil.descendant(self, None, "reference")
        if referenceElement is not None: # formula test suite
            return "{0}#{1}".format(referenceElement.get("specification"), referenceElement.get("id"))
        referenceElement = XmlUtil.descendant(self, None, "documentationReference")
        if referenceElement is not None: # w3c test suite
            return referenceElement.get("{http://www.w3.org/1999/xlink}href")
        descriptionElement = XmlUtil.descendant(self, None, "description")
        if descriptionElement is not None and descriptionElement.get("reference"):
            return descriptionElement.get("reference")  # xdt test suite
        if self.getparent().get("description"):
            return self.getparent().get("description")  # base spec 2.1 test suite
        functRegistryRefElt = XmlUtil.descendant(self.getparent(), None, "reference")
        if functRegistryRefElt is not None: # function registry
            return functRegistryRefElt.get("{http://www.w3.org/1999/xlink}href")
        return None

    @property
    def readMeFirstUris(self):
        try:
            return self._readMeFirstUris
        except AttributeError:
            self._readMeFirstUris = []
            self.readMeFirstElements = []
            # first look if any plugin method to get readme first URIs
            if not any(pluginXbrlMethod(self)
                       for pluginXbrlMethod in pluginClassMethods("ModelTestcaseVariation.ReadMeFirstUris")):
                if self.localName == "testGroup":  #w3c testcase
                    instanceTestElement = XmlUtil.descendant(self, None, "instanceTest")
                    if instanceTestElement is not None: # take instance first
                        self._readMeFirstUris.append(XmlUtil.descendantAttr(instanceTestElement, None,
                                                                            "instanceDocument",
                                                                            "{http://www.w3.org/1999/xlink}href"))
                        self.readMeFirstElements.append(instanceTestElement)
                    else:
                        schemaTestElement = XmlUtil.descendant(self, None, "schemaTest")
                        if schemaTestElement is not None:
                            self._readMeFirstUris.append(XmlUtil.descendantAttr(schemaTestElement, None,
                                                                                "schemaDocument",
                                                                                "{http://www.w3.org/1999/xlink}href"))
                            self.readMeFirstElements.append(schemaTestElement)
                elif self.localName == "test-case":  #xpath testcase
                    inputFileElement = XmlUtil.descendant(self, None, "input-file")
                    if inputFileElement is not None: # take instance first
                        self._readMeFirstUris.append("TestSources/" + inputFileElement.text + ".xml")
                        self.readMeFirstElements.append(inputFileElement)
                elif self.resultIsTaxonomyPackage:
                    self._readMeFirstUris.append( os.path.join(self.modelDocument.filepathdir, "tests", self.get("name") + ".zip") )
                    self.readMeFirstElements.append(self)
                else:
                    # default built-in method for readme first uris
                    for anElement in self.iterdescendants():
                        if isinstance(anElement,ModelObject) and anElement.get("readMeFirst") == "true":
                            if anElement.get("{http://www.w3.org/1999/xlink}href"):
                                uri = anElement.get("{http://www.w3.org/1999/xlink}href")
                            else:
                                uri = XmlUtil.innerText(anElement)
                            if anElement.get("name"):
                                self._readMeFirstUris.append( (ModelValue.qname(anElement, anElement.get("name")), uri) )
                                self.readMeFirstElements.append(anElement)
                            elif anElement.get("dts"):
                                self._readMeFirstUris.append( (anElement.get("dts"), uri) )
                                self.readMeFirstElements.append(anElement)
                            else:
                                self._readMeFirstUris.append(uri)
                                self.readMeFirstElements.append(anElement)
            if not self._readMeFirstUris:  # provide a dummy empty instance document
                self._readMeFirstUris.append(os.path.join(self.modelXbrl.modelManager.cntlr.configDir, "empty-instance.xml"))
                self.readMeFirstElements.append(None)
            return self._readMeFirstUris

    @property
    def dataUris(self):
        try:
            return self._dataUris
        except AttributeError:
            self._dataUris = defaultdict(list) # may contain instances, schemas, linkbases
            for dataElement in XmlUtil.descendants(self, None, ("data", "input")):
                for elt in XmlUtil.descendants(dataElement, None, ("xsd", "schema", "linkbase", "instance")):
                    self._dataUris["schema" if elt.localName == "xsd" else elt.localName].append(elt.textValue.strip())
            return self._dataUris

    @property
    def parameters(self):
        try:
            return self._parameters
        except AttributeError:
            self._parameters = dict([
                (ModelValue.qname(paramElt, paramElt.get("name")), # prefix-less parameter names take default namespace of element
                 (ModelValue.qname(paramElt, paramElt.get("datatype")),paramElt.get("value")))
                for paramElt in XmlUtil.descendants(self, self.namespaceURI, "parameter")])
            return self._parameters

    @property
    def resultIsVersioningReport(self):
        return XmlUtil.descendant(XmlUtil.descendant(self, None, "result"), None, "versioningReport") is not None

    @property
    def versioningReportUri(self):
        return XmlUtil.text(XmlUtil.descendant(self, None, "versioningReport"))

    @property
    def resultIsXbrlInstance(self):
        return XmlUtil.descendant(XmlUtil.descendant(self, None, "result"), None, "instance") is not None

    @property
    def resultXbrlInstanceUri(self):
        for pluginXbrlMethod in pluginClassMethods("ModelTestcaseVariation.ResultXbrlInstanceUri"):
            resultInstanceUri = pluginXbrlMethod(self)
            if resultInstanceUri is not None:
                return resultInstanceUri or None # (empty string returns None)

        for resultElt in self.iterdescendants("{*}result"):
            for instElt in resultElt.iterdescendants("{*}instance"):
                if (instElt.get("target") or "") == (self.ixdsTarget or ""): # match null and emptyString
                    return XmlUtil.text(instElt)
        return None

    @property
    def resultIsInfoset(self):
        if self.modelDocument.outpath:
            result = XmlUtil.descendant(self, None, "result")
            if result is not None:
                return XmlUtil.child(result, None, "file") is not None or XmlUtil.text(result).endswith(".xml")
        return False

    @property
    def resultInfosetUri(self):
        result = XmlUtil.descendant(self, None, "result")
        if result is not None:
            child = XmlUtil.child(result, None, "file")
            return os.path.join(self.modelDocument.outpath, XmlUtil.text(child if child is not None else result))
        return None

    @property
    def resultIsTable(self):
        result = XmlUtil.descendant(self, None, "result")
        if result is not None :
            child = XmlUtil.child(result, None, "table")
            if child is not None and XmlUtil.text(child).endswith(".xml"):
                return True
        return False

    @property
    def resultTableUri(self):
        result = XmlUtil.descendant(self, None, "result")
        if result is not None:
            child = XmlUtil.child(result, None, "table")
            if child is not None:
                return os.path.join(self.modelDocument.outpath, XmlUtil.text(child))
        return None

    @property
    def resultIsTaxonomyPackage(self):
        return any(e.localName for e in XmlUtil.descendants(self,None,TXMY_PKG_SRC_ELTS))

    @property
    def variationDiscoversDTS(self):
        return any(e.localName != "taxonomyPackage" # find any nonTP element (instance, schema, linkbase, etc)
                   for e in self.iterdescendants()
                   if isinstance(e,ModelObject) and e.get("readMeFirst") == "true")

    @property
    def cfcnCall(self):
        # tuple of (expression, element holding the expression)
        try:
            return self._cfcnCall
        except AttributeError:
            self._cfcnCall = None
            if self.localName == "test-case":  #xpath testcase
                queryElement = XmlUtil.descendant(self, None, "query")
                if queryElement is not None:
                    filepath = (self.modelDocument.filepathdir + "/" + "Queries/XQuery/" +
                                self.get("FilePath") + queryElement.get("name") + '.xq')
                    if os.sep != "/": filepath = filepath.replace("/", os.sep)
                    with io.open(filepath, 'rt', encoding='utf-8') as f:
                        self._cfcnCall = (f.read(), self)
            else:
                for callElement in XmlUtil.descendants(self, XbrlConst.cfcn, "call"):
                    self._cfcnCall = (XmlUtil.innerText(callElement), callElement)
                    break
            if self._cfcnCall is None and self.namespaceURI == "http://xbrl.org/2011/conformance-rendering/transforms":
                name = self.getparent().get("name")
                input = self.get("input")
                if name and input:
                    self._cfcnCall =  ("{0}('{1}')".format(name, input.replace("'","''")), self)
            return self._cfcnCall

    @property
    def cfcnTest(self):
        # tuple of (expression, element holding the expression)
        try:
            return self._cfcnTest
        except AttributeError:
            self._cfcnTest = None
            if self.localName == "test-case":  #xpath testcase
                outputFileElement = XmlUtil.descendant(self, None, "output-file")
                if outputFileElement is not None and outputFileElement.get("compare") == "Text":
                    filepath = (self.modelDocument.filepathdir + "/" + "ExpectedTestResults/" +
                                self.get("FilePath") + outputFileElement.text)
                    if os.sep != "/": filepath = filepath.replace("/", os.sep)
                    with io.open(filepath, 'rt', encoding='utf-8') as f:
                        self._cfcnTest = ("xs:string($result) eq '{0}'".format(f.read()), self)
            else:
                testElement = XmlUtil.descendant(self, XbrlConst.cfcn, "test")
                if testElement is not None:
                    self._cfcnTest = (XmlUtil.innerText(testElement), testElement)
                elif self.namespaceURI == "http://xbrl.org/2011/conformance-rendering/transforms":
                    output = self.get("output")
                    if output:
                        self._cfcnTest =  ("$result eq '{0}'".format(output.replace("'","''")), self)
            return self._cfcnTest

    @property
    def expected(self):
        for pluginXbrlMethod in pluginClassMethods("ModelTestcaseVariation.ExpectedResult"):
            expected = pluginXbrlMethod(self)
            if expected:
                return expected
        # default behavior without plugins
        if self.localName == "testcase":
            return self.document.basename[:4]   #starts with PASS or FAIL
        elif self.localName == "testGroup":  #w3c testcase
            instanceTestElement = XmlUtil.descendant(self, None, "instanceTest")
            if instanceTestElement is not None: # take instance first
                return XmlUtil.descendantAttr(instanceTestElement, None, "expected", "validity")
            else:
                schemaTestElement = XmlUtil.descendant(self, None, "schemaTest")
                if schemaTestElement is not None:
                    return XmlUtil.descendantAttr(schemaTestElement, None, "expected", "validity")
        resultElement = XmlUtil.descendant(self, None, "result")
        if resultElement is not None:
            expected = resultElement.get("expected")
            if expected and resultElement.get("nonStandardErrorCodes") == "true":
                # if @expected and @nonStandardErrorCodes then use expected instead of error codes
                return expected
        errorElements = XmlUtil.descendants(self, None, "error")
        resultElement = XmlUtil.descendant(self, None, "result")
        if isinstance(errorElements,list) and len(errorElements) > 0:
            return [ModelValue.qname(e, e.stringValue) if not e.get("nonStandardErrorCodes")
                    else e.stringValue
                    for e in errorElements]
        #else:
        #    errorElement = errorElements[1]
        #    if errorElement is not None and not errorElement.get("nonStandardErrorCodes"):
        #        _errorText = XmlUtil.text(errorElement)
        #        if ' ' in _errorText: # list of tokens
        #            return _errorText
        #        return ModelValue.qname(errorElement, _errorText)  # turn into a QName
        if resultElement is not None:
            if expected:
                return expected
            for assertElement in XmlUtil.children(resultElement, None, "assert"):
                num = assertElement.get("num")
                if num == "99999": # inline test, use name as expected
                    return assertElement.get("name")
                if len(num) == 5:
                    return "EFM.{0}.{1}.{2}".format(num[0],num[1:3],num[3:6])
            asserTests = {}
            for atElt in XmlUtil.children(resultElement, None, "assertionTests"):
                try:
                    asserTests[atElt.get("assertionID")] = (int(atElt.get("countSatisfied")),int(atElt.get("countNotSatisfied")))
                except ValueError:
                    pass
            if asserTests:
                return asserTests
        elif self.get("result"):
            return self.get("result")

        return None

    @property
    def match(self) -> Optional[str]:
        resultElement = XmlUtil.descendant(self, None, "result")
        if resultElement is None:
            return None
        return resultElement.get('match')

    @property
    def expectedCount(self):
        for pluginXbrlMethod in pluginClassMethods("ModelTestcaseVariation.ExpectedCount"):
            _count = pluginXbrlMethod(self)
            if _count is not None: # ignore plug in if not a plug-in-recognized test case
                return _count
        return None


    @property
    def severityLevel(self):
        for pluginXbrlMethod in pluginClassMethods("ModelTestcaseVariation.ExpectedSeverity"):
            severityLevelName = pluginXbrlMethod(self)
            if severityLevelName: # ignore plug in if not a plug-in-recognized test case
                return logging._checkLevel(severityLevelName)
        # default behavior without plugins
        # SEC error cases have <assert severity={err|wrn}>...
        if (XmlUtil.descendant(self, None, "assert", attrName="severity", attrValue="wrn") is not None or
            XmlUtil.descendant(self, None, "result", attrName="severity", attrValue="warning") is not None):
            return logging._checkLevel("WARNING")
        return logging._checkLevel("INCONSISTENCY")

    @property
    def blockedMessageCodes(self):
        blockedCodesRegex = XmlUtil.descendantAttr(self, None, ("results","result"), "blockedMessageCodes") # DQC 4/5 test suite
        if not blockedCodesRegex:
            ignoredCodes = XmlUtil.descendants(self, None, "ignore-error") # ESEF test suite
            if ignoredCodes:
                blockedCodesRegex = "|".join(".*" + c.stringValue for c in ignoredCodes)
        return blockedCodesRegex

    @property
    def expectedVersioningReport(self):
        XmlUtil.text(XmlUtil.text(XmlUtil.descendant(XmlUtil.descendant(self, None, "result"), None, "versioningReport")))

    @property
    def propertyView(self):
        assertions = []
        for assertionElement in XmlUtil.descendants(self, None, "assertionTests"):
            assertions.append(("assertion",assertionElement.get("assertionID")))
            assertions.append(("   satisfied", assertionElement.get("countSatisfied")))
            assertions.append(("   not sat.", assertionElement.get("countNotSatisfied")))
        '''
        for assertionElement in XmlUtil.descendants(self, None, "assert"):
            efmNum = assertionElement.get("num")
            assertions.append(("assertion",
                               "EFM.{0}.{1}.{2}".format(efmNum[0], efmNum[1:2], efmNum[3:4])))
            assertions.append(("   not sat.", "1"))
        '''
        readMeFirsts = [("readFirst", readMeFirstUri) for readMeFirstUri in self.readMeFirstUris]
        parameters = []
        if len(self.parameters) > 0: parameters.append(("parameters", None))
        for pName, pTypeValue in self.parameters.items():
            parameters.append((pName,pTypeValue[1]))
        return [("id", self.id),
                ("name", self.name),
                ("description", self.description)] + \
                readMeFirsts + \
                parameters + \
               [("status", self.status),
                ("call", self.cfcnCall[0]) if self.cfcnCall else (),
                ("test", self.cfcnTest[0]) if self.cfcnTest else (),
                ("infoset", self.resultInfosetUri) if self.resultIsInfoset else (),
                ("expected", self.expected) if self.expected else (),
                ("actual", " ".join(str(i) for i in self.actual) if len(self.actual) > 0 else ())] + \
                assertions

    def __repr__(self):
        return ("modelTestcaseVariation[{0}]{1})".format(self.objectId(),self.propertyView))



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, StringVar, VERTICAL, HORIZONTAL, N, S, E, W, messagebox
try:
    from tkinter.ttk import Frame, Button, Treeview, Scrollbar
except ImportError:
    from ttk import Frame, Button, Treeview, Scrollbar
import os, sys
from collections import defaultdict
import regex as re
from arelle.Cntlr import Cntlr
from arelle.CntlrWinTooltip import ToolTip
from arelle.UrlUtil import isHttpUrl
from arelle.PackageManager import parsePackage
from arelle.PythonUtil import attrdict
from arelle import PluginManager

'''
caller checks accepted, if True, caller retrieves url
'''

ARCHIVE = 1
ENTRY_POINTS = 2
DISCLOSURE_SYSTEM = 3
PLUGIN = 4
PACKAGE = 5

reportIxdsPattern = re.compile(r"^([^/]+/reports/[^/]+)/[^/]+$")

def askArchiveFile(parent, filesource, multiselect=False, selectFiles=None):
    try:
        filenames = filesource.dir
        if filenames is not None:   # an IO or other error can return None
            if filesource.isTaxonomyPackage:
                dialog = DialogOpenArchive(parent,
                                           ENTRY_POINTS,
                                           filesource,
                                           filenames,
                                           _("Select Entry Point"),
                                           _("File"),
                                           showAltViewButton=True,
                                           multiselect=multiselect)
            else:
                dialog = DialogOpenArchive(parent,
                                           ARCHIVE,
                                           filesource,
                                           filenames,
                                           _("Select Archive File"),
                                           _("File"),
                                           multiselect=multiselect,
                                           selectFiles=selectFiles)
            if dialog.accepted:
                return filesource.url
    except Exception as e:
        err = _("Unable to open archive; the underlying error was: {0}").format(e)
        parent.addToLog(err)
        messagebox.showerror(_("Unable to open archive"), err)
    return None

def selectDisclosureSystem(parent, disclosureSystem):

    disclosureSystemSelections = disclosureSystem.dir

    # if no disclosure system to select, user may need to enable applicable plugin(s)
    if not disclosureSystemSelections and messagebox.askokcancel(
        _("Load disclosure systems"),
        _("Disclosure systems are provided by plug-ins, no applicable plug-in(s) have been enabled. \n\n"
          "Press OK to open the plug-in manager and select plug-in(s) (e.g., validate or EdgarRenderer).")):
        from arelle import DialogPluginManager
        DialogPluginManager.dialogPluginManager(parent)
        return None

    dialog = DialogOpenArchive(parent,
                               DISCLOSURE_SYSTEM,
                               disclosureSystem,
                               disclosureSystemSelections,
                               _("Select Disclosure System"),
                               _("Disclosure System"))
    if dialog and dialog.accepted:
        return disclosureSystem.selection
    return None

def selectPlugin(parent, pluginChoices):

    filesource = attrdict(isRss=False, url="Plug-ins", selection="") # emulates a filesource object for the selection return
    dialog = DialogOpenArchive(parent,
                               PLUGIN,
                               filesource,
                               pluginChoices,
                               _("Name"),
                               _("Select Plug-in Module"))
    if dialog and dialog.accepted:
        return filesource.selection
    return None

def selectPackage(parent, packageChoices):

    filesource = attrdict(isRss=False, url="Packages", selection="") # emulates a filesource object for the selection return
    dialog = DialogOpenArchive(parent,
                               PACKAGE,
                               filesource,
                               packageChoices,
                               _("Name"),
                               _("Select Package"))
    if dialog and dialog.accepted:
        return filesource.selection
    return None


class DialogOpenArchive(Toplevel):
    def __init__(self, parent, openType, filesource, filenames, title, colHeader, showAltViewButton=False, multiselect=False, selectFiles=None):
        if isinstance(parent, Cntlr):
            cntlr = parent
            parent = parent.parent # parent is cntlrWinMain
        else: # parent is a Toplevel dialog
            cntlr = parent.cntlr
        super(DialogOpenArchive, self).__init__(parent)
        self.parent = parent
        self.showAltViewButton = showAltViewButton
        self.selectFiles = selectFiles
        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.accepted = False

        self.transient(self.parent)

        frame = Frame(self)

        treeFrame = Frame(frame, width=500)
        vScrollbar = Scrollbar(treeFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(treeFrame, orient=HORIZONTAL)
        self.treeView = Treeview(treeFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set)
        self.treeView.grid(row=0, column=0, sticky=(N, S, E, W))
        self.treeView.config(selectmode="extended" if multiselect else "browse")
        hScrollbar["command"] = self.treeView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.treeView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        treeFrame.columnconfigure(0, weight=1)
        treeFrame.rowconfigure(0, weight=1)
        treeFrame.grid(row=0, column=0, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        self.treeView.focus_set()

        if openType not in (PLUGIN, PACKAGE):
            cntlr.showStatus(_("loading archive {0}").format(filesource.url))
        self.filesource = filesource
        self.filenames = filenames
        self.selection = filesource.selection
        self.hasToolTip = False
        self.multiselect = multiselect
        selectedNode = None

        if openType == ENTRY_POINTS:
            try:
                metadataFiles = filesource.taxonomyPackageMetadataFiles
                ''' take first for now
                if len(metadataFiles) != 1:
                    raise IOError(_("Taxonomy package contained more than one metadata file: {0}.")
                                  .format(', '.join(metadataFiles)))
                '''
                metadataFile = metadataFiles[0]
                metadata = filesource.basefile + os.sep + metadataFile
                self.metadataFilePrefix = os.sep.join(os.path.split(metadataFile)[:-1])
                if self.metadataFilePrefix:
                    self.metadataFilePrefix += "/"  # zip contents have /, never \ file seps
                self.taxonomyPkgMetaInf = '{}/META-INF/'.format(
                            os.path.splitext(os.path.basename(filesource.url))[0])


                self.taxonomyPackage = parsePackage(cntlr, filesource, metadata,
                                                    os.sep.join(os.path.split(metadata)[:-1]) + os.sep)


                if self.taxonomyPackage["entryPoints"]:
                    # may have instance documents too
                    self.packageContainedInstances = []
                    self.packageContainedIXDSes = defaultdict(list)
                    packageContentInstanceCounts = {}
                    packageContentTypeCounts = {}
                    for suffix in (".xhtml", ".htm", ".html", ".xbrl"): # try for suffixes in order of likelihood to have instance
                        for potentialInstance in filesource.dir:
                            if potentialInstance.endswith(suffix):
                                m = reportIxdsPattern.match(potentialInstance) # IXDS
                                if multiselect and m: # only in ixds multiselect mode
                                    _type = "Inline Doc Set"
                                    self.packageContainedIXDSes[m.group(1)].append(potentialInstance)
                                    potentialInstance = m.group(1) # use package name only
                                elif suffix == ".xbrl":
                                    _type = "xBRL-XML Instance"
                                else:
                                    _type = "Inline Instance"
                                if not self.packageContainedInstances or self.packageContainedInstances[-1][0] != potentialInstance:
                                    self.packageContainedInstances.append([potentialInstance, _type])
                                    packageContentInstanceCounts[potentialInstance] = packageContentInstanceCounts.get(potentialInstance, 0) + 1
                                    packageContentTypeCounts[_type] = packageContentTypeCounts.get(_type, 0) + 1
                        if self.packageContainedInstances:
                            break
                    if self.packageContainedInstances: # add sequences to any duplicated entry types
                        for _type, count in packageContentTypeCounts.items():
                            if count > 1:
                                _dupNo = 0
                                for i in range(len(self.packageContainedInstances)):
                                    if self.packageContainedInstances[i][1] == _type:
                                        _dupNo += 1
                                        self.packageContainedInstances[i][1] = "{} {}".format(_type, _dupNo)
                        for _instance, count in packageContentInstanceCounts.items():
                            if count > 1:
                                _dupNo = 0
                                for i in range(len(self.packageContainedInstances)):
                                    if self.packageContainedInstances[i][0] == _instance:
                                        _dupNo += 1
                                        self.packageContainedInstances[i][0] = "{} {}".format(_instance, _dupNo)

                else:
                    # may be a catalog file with no entry oint names
                    openType = ARCHIVE  # no entry points to show, just archive
                    self.showAltViewButton = False
            except Exception as e:
                self.close()
                err = _("Failed to parse metadata; the underlying error was: {0}").format(e)
                messagebox.showerror(_("Malformed taxonomy package"), err)
                cntlr.addToLog(err)
                return

        if openType not in (PLUGIN, PACKAGE):
            cntlr.showStatus(None)

        if openType in (DISCLOSURE_SYSTEM, PLUGIN, PACKAGE):
            y = 3
        else:
            y = 1

        okButton = Button(frame, text=_("OK"), command=self.ok)
        cancelButton = Button(frame, text=_("Cancel"), command=self.close)
        okButton.grid(row=y, column=2, sticky=(S,E,W), pady=3)
        cancelButton.grid(row=y, column=3, sticky=(S,E,W), pady=3, padx=3)

        if self.showAltViewButton:
            self.altViewButton = Button(frame, command=self.showAltView)
            self.altViewButton.grid(row=y, column=0, sticky=(S,W), pady=3, padx=3)

        self.loadTreeView(openType, colHeader, title)

        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))
        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        window.rowconfigure(0, weight=1)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.close)

        self.toolTipText = StringVar()
        if self.hasToolTip:
            self.treeView.bind("<Motion>", self.motion, '+')
            self.treeView.bind("<Leave>", self.leave, '+')
            self.toolTipText = StringVar()
            self.toolTip = ToolTip(self.treeView,
                                   textvariable=self.toolTipText,
                                   wraplength=640,
                                   follow_mouse=True,
                                   state="disabled")
            self.toolTipRowId = None

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()

        self.wait_window(self)




    def loadTreeView(self, openType, title, colHeader):
        self.title(title)
        self.openType = openType
        selectedNodes = []

        # clear previous treeview entries
        for previousNode in self.treeView.get_children(""):
            self.treeView.delete(previousNode)

        # set up treeView widget and tabbed pane
        if openType in (ARCHIVE, DISCLOSURE_SYSTEM, PLUGIN, PACKAGE):
            if openType in (PLUGIN, PACKAGE): width = 770
            else: width = 500
            self.treeView.column("#0", width=width, anchor="w")
            self.treeView.heading("#0", text=colHeader)
            self.isRss = getattr(self.filesource, "isRss", False)
            if self.isRss:
                self.treeView.column("#0", width=350, anchor="w")
                self.treeView["columns"] = ("descr", "date", "instDoc")
                self.treeView.column("descr", width=50, anchor="center", stretch=False)
                self.treeView.heading("descr", text="Form")
                self.treeView.column("date", width=170, anchor="w", stretch=False)
                self.treeView.heading("date", text="Pub Date")
                self.treeView.column("instDoc", width=200, anchor="w", stretch=False)
                self.treeView.heading("instDoc", text="Instance Document")
            elif openType == PLUGIN:
                self.treeView.column("#0", width=250, anchor="w")
                self.treeView["columns"] = ("vers", "descr", "license")
                self.treeView.column("vers", width=60, anchor="w", stretch=False)
                self.treeView.heading("vers", text="Version")
                self.treeView.column("descr", width=300, anchor="w", stretch=False)
                self.treeView.heading("descr", text="Description")
                self.treeView.column("license", width=60, anchor="w", stretch=False)
                self.treeView.heading("license", text="License")
            elif openType == PACKAGE:
                self.treeView.column("#0", width=200, anchor="w")
                self.treeView["columns"] = ("vers", "descr", "license")
                self.treeView.column("vers", width=100, anchor="w", stretch=False)
                self.treeView.heading("vers", text="Version")
                self.treeView.column("descr", width=400, anchor="w", stretch=False)
                self.treeView.heading("descr", text="Description")
                self.treeView.column("license", width=70, anchor="w", stretch=False)
                self.treeView.heading("license", text="License")
            else:
                self.treeView["columns"] = tuple()

            loadedPaths = []
            for i, filename in enumerate(self.filenames):
                if isinstance(filename,tuple):
                    if self.isRss:
                        form, date, instDoc = filename[2:5]
                    elif openType == PLUGIN:
                        name, vers, descr, license = filename[3:7]
                    elif openType == PACKAGE:
                        vers, descr, license = filename[3:6]
                    filename = filename[0] # ignore tooltip
                    self.hasToolTip = True
                if filename.endswith("/"):
                    filename = filename[:-1]
                path = filename.split("/")
                if not self.isRss and len(path) > 1 and path[:-1] in loadedPaths:
                    parent = "file{0}".format(loadedPaths.index(path[:-1]))
                else:
                    parent = ""
                node = self.treeView.insert(parent, "end", "file{0}".format(i), text=path[-1])
                if self.isRss:
                    self.treeView.set(node, "descr", form)
                    self.treeView.set(node, "date", date)
                    self.treeView.set(node, "instDoc", os.path.basename(instDoc))
                elif openType == PLUGIN:
                    self.treeView.set(node, "vers", vers)
                    self.treeView.set(node, "descr", descr)
                    self.treeView.set(node, "license", license)
                elif openType == PACKAGE:
                    self.treeView.set(node, "vers", vers)
                    self.treeView.set(node, "descr", descr)
                    self.treeView.set(node, "license", license)
                if self.selectFiles:
                    if filename in self.selectFiles:
                        selectedNodes.append(node)
                elif self.selection == filename:
                    selectedNodes.append(node)
                loadedPaths.append(path)

        elif openType == ENTRY_POINTS:
            self.treeView.column("#0", width=200, anchor="w")
            self.treeView.heading("#0", text="Name")

            self.treeView["columns"] = ("url",)
            self.treeView.column("url", width=300, anchor="w")
            self.treeView.heading("url", text="URL")

            for fileType, fileUrl in getattr(self, "packageContainedInstances", ()):
                self.treeView.insert("", "end", fileUrl,
                                     values=fileType,
                                     text=fileUrl or urls[0][2])
            for name, urls in sorted(self.taxonomyPackage["entryPoints"].items(), key=lambda i:i[0][2]):
                self.treeView.insert("", "end", name,
                                     values="\n".join(url[1] for url in urls),
                                     text=name or urls[0][2])

            self.hasToolTip = True
        else: # unknown openType
            return None
        if selectedNodes:
            self.treeView.see(selectedNodes[0])
            self.treeView.selection_set(selectedNodes)

        if self.showAltViewButton:
            self.altViewButton.config(text=_("Show Files") if openType == ENTRY_POINTS else _("Show Entries"))


    def ok(self, event=None):
        selection = self.treeView.selection()
        if len(selection) > 0:
            filename = None
            if self.openType in (ARCHIVE, DISCLOSURE_SYSTEM):
                if self.multiselect:
                    filenames = []
                for _selection in selection:
                    filename = self.filenames[int(_selection[4:])]
                    if isinstance(filename,tuple):
                        if self.isRss:
                            filename = filename[4]
                        else:
                            filename = filename[0]
                    if self.multiselect:
                        filenames.append(filename)
                    else:
                        break
                if self.multiselect and filenames:
                    self.filesource.select(filenames) # array of file names
                    self.accepted = True
                    self.close()
            elif self.openType == ENTRY_POINTS:
                epName = selection[0]
                #index 0 is the remapped Url, as opposed to the canonical one used for display
                # Greg Acsone reports [0] does not work for Corep 1.6 pkgs, need [1], old style packages
                filenames = []
                for _url, _type in self.packageContainedInstances: # check if selection was an inline instance
                    if _type in selection:
                        if _url in self.packageContainedIXDSes: # taxonomy package
                            filenames.extend(self.packageContainedIXDSes[_url])
                        else: # single instance
                            filenames.append(_url)
                if not filenames: # else if it's a named taxonomy entry point of an installed package
                    for url in self.taxonomyPackage["entryPoints"][epName]:
                        filename = url[1] # use unmapped file name
                        if not filename.endswith("/"):
                            # check if it's an absolute URL rather than a path into the archive
                            if not isHttpUrl(filename) and self.metadataFilePrefix != self.taxonomyPkgMetaInf:
                                # assume it's a path inside the archive:
                                filename = self.metadataFilePrefix + filename
                        filenames.append(filename)
                if filenames:
                    self.filesource.select(filenames)
                    self.accepted = True
                    self.close()
                return
            elif self.openType in (PLUGIN, PACKAGE):
                filename = self.filenames[int(selection[0][4:])][2]
            if filename is not None and not self.multiselect and not filename.endswith("/"):
                if hasattr(self, "taxonomyPackage"):
                    # attempt to unmap the filename to original file
                    # will be mapped again in loading, but this allows schemaLocation to be unmapped
                    for prefix, remapping in self.taxonomyPackage["remappings"].items():
                        if isHttpUrl(remapping):
                            remapStart = remapping
                        else:
                            remapStart = self.metadataFilePrefix + remapping
                        if filename.startswith(remapStart):
                            # set unmmapped file
                            filename = prefix + filename[len(remapStart):]
                            break
                        if (self.metadataFilePrefix.endswith("/META-INF/") and isHttpUrl(prefix) and
                            filename.startswith(self.metadataFilePrefix[:-10]) and
                            filename.startswith(remapping[len(self.filesource.url)+1:])):
                            # recover unmapped file name for chosen in-archive relative file
                            filename = prefix + filename[len(remapping) - len(self.filesource.url) - 1:]
                if self.openType in (PLUGIN, PACKAGE):
                    self.filesource.selection = filename
                else:
                    self.filesource.select(filename)
                self.accepted = True
                self.close()


    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()

    def showAltView(self, event=None):
        if self.openType == ENTRY_POINTS:
            self.loadTreeView(ARCHIVE, _("Select Entry Point"), _("File"))
        else:
            self.loadTreeView(ENTRY_POINTS, _("Select Archive File"), _("File"))

    def leave(self, *args):
        self.toolTipRowId = None

    def motion(self, *args):
        tvRowId = self.treeView.identify_row(args[0].y)
        if tvRowId != self.toolTipRowId:
            text = None
            if self.openType in (ARCHIVE, DISCLOSURE_SYSTEM, PLUGIN, PACKAGE):
                self.toolTipRowId = tvRowId
                if tvRowId and len(tvRowId) > 4:
                    try:
                        text = self.filenames[ int(tvRowId[4:]) ]
                        if isinstance(text, tuple):
                            text = (text[1] or "").replace("\\n","\n")
                    except (KeyError, ValueError):
                        pass
            elif self.openType == ENTRY_POINTS:
                try:
                    text = "{0}\n{1}".format(tvRowId,
                             "\n".join(url[1] for url in self.taxonomyPackage["entryPoints"][tvRowId]))
                except KeyError:
                    pass
            self.setToolTip(text)

    def setToolTip(self, text):
        self.toolTip._hide()
        if text:
            self.toolTipText.set(text)
            self.toolTip.configure(state="normal")
            self.toolTip._schedule()
        else:
            self.toolTipText.set("")
            self.toolTip.configure(state="disabled")



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from arelle.ModelValue import qname
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText

NAMESPACE_ARR = 'http://xbrl.dcca.dk/arr'
NAMESPACE_CMN = 'http://xbrl.dcca.dk/cmn'
NAMESPACE_FSA = 'http://xbrl.dcca.dk/fsa'
NAMESPACE_GSD = 'http://xbrl.dcca.dk/gsd'
NAMESPACE_SOB = 'http://xbrl.dcca.dk/sob'


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self, validateXbrl: ValidateXbrl) -> PluginValidationDataExtension:
        return PluginValidationDataExtension(
            self.name,
            annualReportTypes=frozenset([
                'Årsrapport',
                'årsrapport',
                'Annual report'
            ]),
            assetsQn=qname(f'{{{NAMESPACE_FSA}}}Assets'),
            classOfReportingEntityQn=qname(f'{{{NAMESPACE_FSA}}}ClassOfReportingEntity'),
            consolidatedMemberQn=qname(f'{{{NAMESPACE_CMN}}}ConsolidatedMember'),
            consolidatedSoloDimensionQn=qname(f'{{{NAMESPACE_CMN}}}ConsolidatedSoloDimension'),
            dateOfApprovalOfAnnualReportQn=qname(f'{{{NAMESPACE_SOB}}}DateOfApprovalOfAnnualReport'),
            dateOfExtraordinaryDividendDistributedAfterEndOfReportingPeriod=
            qname(f'{{{NAMESPACE_FSA}}}DateOfExtraordinaryDividendDistributedAfterEndOfReportingPeriod'),
            dateOfGeneralMeetingQn=qname(f'{{{NAMESPACE_GSD}}}DateOfGeneralMeeting'),
            descriptionOfQualificationsOfAssuranceEngagementPerformedQn=qname(f'{{{NAMESPACE_ARR}}}DescriptionOfQualificationsOfAssuranceEngagementPerformed'),
            distributionOfResultsQns=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}DistributionsResultDistribution'),
                qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryDistributions'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedDividendRecognisedInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedExtraordinaryDividendRecognisedInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedExtraordinaryDividendRecognisedInLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToHedgeFund'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToReserveFund'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToReservesAvailable'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromEquityAttributableToParent'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromMinorityInterests'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromOtherStatutoryReserves'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveAccordingToArticlesOfAssociation'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveForNetRevaluationAccordingToEquityMethod'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveForNetRevaluationOfInvestmentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromRestOfOtherReserves'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromRetainedEarnings'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForCurrentValueAdjustmentsOfCurrencyGains'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForCurrentValueOfHedging'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForDevelopmentExpenditure'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForEntrepreneurialCompany'),
            ]),
            equityQn=qname(f'{{{NAMESPACE_FSA}}}Equity'),
            extraordinaryCostsQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryCosts'),
            extraordinaryIncomeQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryIncome'),
            extraordinaryResultBeforeTaxQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryResultBeforeTax'),
            fr37RestrictedText='has not given rise to reservations',
            informationOnTypeOfSubmittedReportQn=qname(f'{{{NAMESPACE_GSD}}}InformationOnTypeOfSubmittedReport'),
            liabilitiesQn=qname(f'{{{NAMESPACE_FSA}}}LiabilitiesAndEquity'),
            positiveProfitThreshold=1000,
            precedingReportingPeriodEndDateQn=qname(f'{{{NAMESPACE_GSD}}}PredingReportingPeriodEndDate'),  # Typo in taxonomy
            precedingReportingPeriodStartDateQn=qname(f'{{{NAMESPACE_GSD}}}PrecedingReportingPeriodStartDate'),
            profitLossQn=qname(f'{{{NAMESPACE_FSA}}}ProfitLoss'),
            proposedDividendRecognisedInEquityQn=qname(f'{{{NAMESPACE_FSA}}}ProposedDividendRecognisedInEquity'),
            reportingPeriodEndDateQn=qname(f'{{{NAMESPACE_GSD}}}ReportingPeriodEndDate'),
            reportingPeriodStartDateQn=qname(f'{{{NAMESPACE_GSD}}}ReportingPeriodStartDate'),
            taxExpenseOnOrdinaryActivitiesQn=qname(f'{{{NAMESPACE_FSA}}}TaxExpenseOnOrdinaryActivities'),
            taxExpenseQn=qname(f'{{{NAMESPACE_FSA}}}TaxExpense'),
            typeOfReportingPeriodDimensionQn=qname(f'{{{NAMESPACE_GSD}}}TypeOfReportingPeriodDimension'),
        )



'''
@author: Acsone S. A.
See COPYRIGHT.md for copyright information.
'''
from tkinter import *
try:
    from tkinter.ttk import *
except ImportError:
    from ttk import *
from arelle.CntlrWinTooltip import ToolTip

class ViewPane:
    def __init__(self, modelXbrl, tabWin, tabTitle,
                 contentView, hasToolTip=False, lang=None):
        self.blockViewModelObject = 0
        self.tabWin = tabWin

        self.viewFrame = contentView
        self.viewFrame.view = self

        tabWin.add(self.viewFrame,text=tabTitle)
        self.modelXbrl = modelXbrl
        self.hasToolTip = hasToolTip
        self.toolTipText = StringVar()
        if hasToolTip:
            self.toolTipText = StringVar()
            self.toolTip = ToolTip(self.gridBody,
                                   textvariable=self.toolTipText,
                                   wraplength=480,
                                   follow_mouse=True,
                                   state="disabled")
            self.toolTipColId = None
            self.toolTipRowId = None
        self.modelXbrl = modelXbrl
        modelManager = self.modelXbrl.modelManager
        self.contextMenuClick = modelManager.cntlr.contextMenuClick
        self.lang = lang
        if modelXbrl:
            modelXbrl.views.append(self)
            if not lang:
                self.lang = modelXbrl.modelManager.defaultLang

    def close(self):
        del self.viewFrame.view
        self.tabWin.forget(self.viewFrame)
        if self in self.modelXbrl.views:
            self.modelXbrl.views.remove(self)
        self.modelXbrl = None

    def select(self):
        self.tabWin.select(self.viewFrame)

    def onClick(self, *args):
        if self.modelXbrl:
            self.modelXbrl.modelManager.cntlr.currentView = self

    def leave(self, *args):
        self.toolTipColId = None
        self.toolTipRowId = None

    def motion(self, *args):
        pass


    def contextMenu(self):
        try:
            return self.menu
        except AttributeError:
            self.menu = Menu( self.viewFrame, tearoff = 0 )
            return self.menu

    def bindContextMenu(self, widget):
        if not widget.bind(self.contextMenuClick):
            widget.bind( self.contextMenuClick, self.popUpMenu )

    def popUpMenu(self, event):
        self.menu.post( event.x_root, event.y_root )

    def menuAddLangs(self):
        langsMenu = Menu(self.viewFrame, tearoff=0)
        self.menu.add_cascade(label=_("Language"), menu=langsMenu, underline=0)
        for lang in sorted(self.modelXbrl.langs):
            langsMenu.add_cascade(label=lang, underline=0,
                                  command=lambda l=lang: self.setLang(l))

    def setLang(self, lang):
        self.lang = lang
        self.view()



'''
See COPYRIGHT.md for copyright information.
'''
import regex as re

def checkForBOMs(modelXbrl, file, mappedUri, filepath, *args, **kwargs):
    # callback is for all opened docs, must only process when SBRNL validation active
    if (modelXbrl.modelManager.validateDisclosureSystem and
        # corrected merge of pre-plugin code per LOGIUS
        getattr(modelXbrl.modelManager.validateDisclosureSystem, "SBRNLplugin", False)):
        #must read file in binary and return nothing to not replace standard loading
        with open(filepath, 'rb') as fb:
            startingBytes = fb.read(8)
            if re.match(b"\\x00\\x00\\xFE\\xFF|\\xFF\\xFE\\x00\\x00|\\x2B\\x2F\\x76\\x38|\\x2B\\x2F\\x76\\x39|\\x2B\\x2F\\x76\\x2B|\\x2B\\x2F\\x76\\x2F|\\xDD\\x73\\x66\\x73|\\xEF\\xBB\\xBF|\\x0E\\xFE\\xFF|\\xFB\\xEE\\x28|\\xFE\\xFF|\\xFF\\xFE",
                        startingBytes):
                modelXbrl.error("SBR.NL.2.1.0.09",
                    _("File MUST not start with a Byte Order Mark (BOM): %(filename)s"),
                    modelObject=modelXbrl, filename=mappedUri)
    return None # must return None for regular document loading to continue



from __future__ import annotations

import logging
import os
import urllib.request
from pathlib import Path

from arelle.RuntimeOptions import RuntimeOptions
# include import start
from arelle.api.Session import Session
from arelle.logging.handlers.StructuredMessageLogHandler import StructuredMessageLogHandler
# include import end
from tests.integration_tests.integration_test_util import get_s3_uri
from tests.integration_tests.scripts.script_util import parse_args, validate_log_xml, assert_result

errors = []
this_file = Path(__file__)
args = parse_args(
    this_file.stem,
    "Confirm ixbrl-viewer plugin runs successfully using Arelle's Python API.",
    arelle=False,
    cache='ixbrl-viewer_cli.zip',
    cache_version_id='P.uruiqpYrdNHGzX.XuJPGS3QS6_qY9g',
)
arelle_offline = args.offline
working_directory = Path(args.working_directory)
test_directory = Path(args.test_directory)
samples_zip_path = test_directory / 'samples.zip'
target_path = samples_zip_path / "samples/src/ixds-test/document1.html"
viewer_path = test_directory / "viewer.html"
report_zip_url = get_s3_uri(
    'ci/packages/IXBRLViewerSamples.zip',
    version_id='6eS7qUUoWLeM9JSSTXOfANkHoLz1Zv5o'
)

print(f"Downloading samples: {samples_zip_path}")
urllib.request.urlretrieve(report_zip_url, samples_zip_path)


class TestFilter(logging.Filter):
    filtered_records = []

    def filter(self, record):
        if record.levelname == 'INFO':
            self.filtered_records.append(record)
            return False
        return True


log_filter = TestFilter()
log_handler = StructuredMessageLogHandler()
print(f"Generating IXBRL viewer: {viewer_path}")
# include start
with open(samples_zip_path, 'rb') as stream:
    options = RuntimeOptions(
        entrypointFile=str(target_path),
        internetConnectivity='offline' if arelle_offline else 'online',
        keepOpen=True,
        logFormat="[%(messageCode)s] %(message)s - %(file)s",
        logPropagate=False,
        pluginOptions={
            'saveViewerDest': str(viewer_path),
            'viewer_feature_review': True,
        },
        plugins='ixbrl-viewer',
    )
    # Plugin default options haven't been applied yet.
    assert not hasattr(options, 'viewerURL')
    with Session() as session:
        session.run(
            options,
            sourceZipStream=stream,
            logHandler=log_handler,
            logFilters=[log_filter],
        )
        # Plugin default options were applied.
        assert hasattr(options, "viewerURL")
        assert options.viewerURL.endswith("ixbrlviewer.js")
        log_xml = session.get_logs('xml')
# include end

print(f"Checking for viewer: {viewer_path}")
if not viewer_path.exists():
    errors.append(f'Viewer not generated at "{viewer_path}"')

print(f"Checking for filtered logs...")
expected_filtered = 5
actual_filtered = len(log_filter.filtered_records)
if actual_filtered != expected_filtered:
    errors.append(f'Expected {expected_filtered} filtered log records, found {actual_filtered}.')

print(f"Checking log XML for errors...")
assert log_xml == log_handler.getXml(clearLogBuffer=False, includeDeclaration=False)
errors += validate_log_xml(log_xml)

assert_result(errors)

print("Cleaning up")
try:
    os.unlink(working_directory / 'python_api_ixbrl-viewer' / 'samples.zip')
    os.unlink(working_directory / 'python_api_ixbrl-viewer' / 'viewer.html')
    os.unlink(working_directory / 'python_api_ixbrl-viewer' / 'ixbrlviewer.js')
except PermissionError as exc:
    print(f"Failed to cleanup test files: {exc}")



from __future__ import annotations

import os
import urllib.request
import zipfile

from pathlib import Path
from shutil import rmtree

from tests.integration_tests.integration_test_util import get_s3_uri
from tests.integration_tests.scripts.script_util import run_arelle, parse_args, validate_log_file, assert_result, prepare_logfile

errors = []
this_file = Path(__file__)
args = parse_args(
    this_file.stem,
    "Confirm ixbrl-viewer plugin runs successfully from the command line.",
    cache=this_file.with_suffix(".zip").name,
    cache_version_id='P.uruiqpYrdNHGzX.XuJPGS3QS6_qY9g',
)
arelle_command = args.arelle
arelle_offline = args.offline
working_directory = Path(args.working_directory)
test_directory = Path(args.test_directory)
arelle_log_file = prepare_logfile(test_directory, this_file)
samples_zip_path = test_directory / 'samples.zip'
samples_directory = test_directory / 'samples'
target_path = samples_directory / "samples/src/ixds-test/document1.html"
viewer_path = test_directory / "viewer.html"
report_zip_url = get_s3_uri(
    'ci/packages/IXBRLViewerSamples.zip',
    version_id='6eS7qUUoWLeM9JSSTXOfANkHoLz1Zv5o'
)

print(f"Downloading samples: {samples_zip_path}")
urllib.request.urlretrieve(report_zip_url, samples_zip_path)

print(f"Extracting samples: {samples_directory}")
with zipfile.ZipFile(samples_zip_path, "r") as zip_ref:
    zip_ref.extractall(samples_directory)

print(f"Generating IXBRL viewer: {viewer_path}")
run_arelle(
    arelle_command,
    plugins=["ixbrl-viewer"],
    additional_args=[
        "-f", str(target_path),
        "--save-viewer", str(viewer_path),
    ],
    offline=arelle_offline,
    logFile=arelle_log_file,
)

print(f"Checking for viewer: {viewer_path}")
if not viewer_path.exists():
    errors.append(f'Viewer not generated at "{viewer_path}"')

print(f"Checking for log errors: {arelle_log_file}")
errors += validate_log_file(arelle_log_file)

assert_result(errors)

print("Cleaning up")
rmtree(working_directory / 'ixbrl-viewer_cli' / 'samples')
os.unlink(working_directory / 'ixbrl-viewer_cli' / 'samples.zip')
os.unlink(working_directory / 'ixbrl-viewer_cli' / 'viewer.html')
os.unlink(working_directory / 'ixbrl-viewer_cli' / 'ixbrlviewer.js')



import csv
import json
import os
from arelle import XbrlConst
from .MessageNumericId import deiSubTblCodes, ftSubTbl, ftSumTbl, ftOfferingTbl, ftOffsetTbl, ftCmbPrsTbl, ft424iTbl, ftStart, ftTableStartCode, ftValidations, ftRuleCode, efmStart
from .Consts import attachmentDocumentTypeValidationRulesFiles
from .FtJsonToOimJson import getLatestTaxonomyFamily


class FtValidations:
    RULES = {
        "Offering": {
            "Rule457aFlg": "Rule 457(a)",
            "Rule457fFlg": "Rule 457(f)",
            "Rule457oFlg": "Rule 457(o)",
            "Rule457uFlg": "Rule 457(u)",
            "Rule457rFlg": "Rule 457(r)",
            "Rule457sFlg": "Rule 457(s)",
            "Rule011Flg": "Rule 0-11",
            "FeesOthrRuleFlg": "Other Rule",
            "Rule415a6Flg": "Rule 415(a)(6)",
            "GnlInstrIIhiFlg": "General Instruction II.H,I"
        },
        "Offset": {
            "Rule457bOffsetFlg": "Rule 457(b)",
            "Rule011a2OffsetFlg": "Rule 0-11(a)(2)",
            "Rule457pOffsetFlg": "Rule 457(p)"
        },
        "Combined Prospectus": {
            "Rule429CmbndPrspctsFlg": "Rule 429"
        }
    }

    def __init__(self, cntlr, typePattern, outName=None):
        self.cntlr = cntlr
        self.typePattern = typePattern
        self.outName = outName
        self._setup()
        return None

    def generateValidations(self):
        validations = sorted(self.parseValidations(), key=lambda x: x.get("Arelle Validation Rule ID", 0) )
        with open(self.outName, "w", newline="") as outFile:
            writer = csv.DictWriter(outFile, fieldnames=self.columns)
            writer.writeheader()
            for validation in validations:
                writer.writerow(validation)

    def parseValidations(self):
        formTypeSevs = []
        for sev in self.validationsJson["sub-type-element-validations"]:
            if sev.get("efm") == "6.5.20":
                formTypeSevs.append(sev)
            else:
                for validation in self.getValidationsFromSev(sev):
                    yield validation
        yield self._getFormTypeSevsValidation(formTypeSevs)

    def getValidationsFromSev(self, sev):
        if sev.get("validation"):
            xbrlNames = self._getxbrlNames(sev)
            for xbrlName in xbrlNames:
                ruleReferences = self._getRuleReference(xbrlName, sev)
                for ruleReference in ruleReferences:
                    for numericID, subTypes in self._getMessageNumericID(xbrlName, sev, ruleReference):
                        row = {}
                        row["Rule Reference"] = ruleReference
                        row["Data Element"] = self._getDataElement(xbrlName, sev)
                        validationLineType = self._getValidationLineType(xbrlName, sev)
                        row["Fees to be Paid"] = validationLineType["Fees to be Paid"]
                        row["Fees Previously Paid"] = validationLineType["Fees Previously Paid"]
                        row["Offset Claim"] = validationLineType["Offset Claim"]
                        row["Offset Source"] = validationLineType["Offset Source"]
                        row["XBRL Name"] = xbrlName
                        row["Arelle Validation rule name"] = sev.get("efm")
                        row["Arelle Validation Rule ID"] = numericID
                        row["Validation Severity"] = self._getValidationSeverity(sev)
                        row["Validation Description"] = sev.get("comment-validation")
                        row["Validation Message"] = self._getValidationMessage(xbrlName, sev)
                        row["Exception Handling"] = sev.get("comment-exception-handling")
                        row["Submission Types"] = subTypes
                        row["Conversion Requested"] = sev.get("comment-conversion-requested")
                        row["Requesting D/O"] = sev.get("comment-requesting-do")
                        row["Date Requested"] = sev.get("comment-date-requested")
                        row["Reason for Conversion"] =sev.get("comment-reason")
                        yield row

    def _setup(self):
        self.validationFileName = self._getValidationsFileName()
        self.validationFilePath = os.path.join(os.path.dirname(__file__), "resources", self.validationFileName)
        self.validationFileBasePath = os.path.dirname(self.validationFilePath)
        self.columns = [
            "Rule Reference", "Data Element", "Fees to be Paid", "Fees Previously Paid", "Offset Claim", "Offset Source",
            "XBRL Name", "Arelle Validation rule name", "Arelle Validation Rule ID", "Validation Severity",
            "Validation Description", "Validation Message", "Exception Handling", "Submission Types", "Conversion Requested",
            "Requesting D/O", "Date Requested", "Reason for Conversion"
            ]
        with open(self.validationFilePath, "r") as file:
            self.validationsJson = json.load(file)
        self.submissionTypeClasses = self._compileSubmissionTypeClasses()
        if not self.outName:
            self.outName = os.path.join(self.validationFileBasePath, "ft-validations.csv")
        latestFFD = getLatestTaxonomyFamily(self.cntlr, "FFD")
        self.cntlr.modelManager.load(latestFFD.href)

    def _getFormTypeSevsValidation(self, formTypeSevs):
        subTypes = self._getSubtypeFormMapping(formTypeSevs)
        for validation in self.getValidationsFromSev(formTypeSevs[0]):
            validation["Submission Types"] = json.dumps(subTypes)
            return validation

    def _getSubtypeFormMapping(self, formTypeSevs):
        mapping = {}
        for formTypeSev in formTypeSevs:
            for subType in formTypeSev.get("sub-types", []):
                if subType in mapping:
                    mapping[subType].extend(formTypeSev.get("value", []))
                else:
                    mapping[subType] = formTypeSev.get("value", [])
        return mapping

    def _compileSubmissionTypeClasses(self):
        STC = self.validationsJson["sub-type-classes"]
        expandedSTC = {}

        def compileSubs(subTypes):
            for subType in subTypes:
                if subType.startswith("@"):
                    for st in compileSubs(STC[subType[1:]]):
                        yield st
                else:
                    yield subType

        for stcName, subTypes in self.validationsJson["sub-type-classes"].items():
            if not "comment" in stcName and not "Comment" in stcName:
                expandedSubTypes = []
                for subType in compileSubs(subTypes):
                    expandedSubTypes.append(subType)
                expandedSTC[stcName] = expandedSubTypes
        return expandedSTC

    def _getMessageFromSev(self, sev):
        if sev.get("message"):
            messageID = sev["message"]
            messageTemplate = self.validationsJson["messages"][messageID]
            return (messageID, messageTemplate)
        if sev.get("validation") and not sev.get("validation") in {"fo", "skip-if-absent"}: # optional validation does not result in any message
            messageID = self.validationsJson["validations"][sev["validation"]]["message"]
            messageTemplate = self.validationsJson["messages"][messageID]
            return (messageID, messageTemplate)
        return (None, None)

    def _getValidationsFileName(self):
        for typePattern, fileName in attachmentDocumentTypeValidationRulesFiles:
            if typePattern == self.typePattern:
                return fileName
        return None

    def _getxbrlNames(self, sev):
        messageID, messageTemplate = self._getMessageFromSev(sev)
        if messageID:
            if sev.get("validation") == "fdepflag-any" and len(self._getAxesFromSev(sev)) > 1:
                return sev["references"][:1]
            if sev.get("validation") == "of-rule":
                return [", ".join(sev["xbrl-names"] if isinstance(sev["xbrl-names"], list) else [sev["xbrl-names"]])]
            if sev.get("axis", sev.get("references-axes")) != sev.get("references-axes", sev.get("axis")):
                # assuming mismatching references-axes would indicate the main axis is being checked.
                return sev["xbrl-names"] if isinstance(sev["xbrl-names"], list) else [sev["xbrl-names"]]
            if sev.get("validation") == "noDups":
                return [f"{tableName} [table]" for tableName in self._getTableNamesFromSev(sev)]
            if "{tag}" in messageID or "{tags}" in messageID:
                return sev["xbrl-names"] if isinstance(sev["xbrl-names"], list) else [sev["xbrl-names"]]
            elif "{otherTag}" in messageID:
                return sev["references"] if isinstance(sev["references"], list) else [sev["references"]]
            elif sev.get("msgCoda"):
                if "{tag}" in sev.get("msgCoda"):
                    return sev["xbrl-names"] if isinstance(sev["xbrl-names"], list) else [sev["xbrl-names"]]
                elif "{otherTag}" in sev.get("msgCoda"):
                    return sev["references"] if isinstance(sev["references"], list) else [sev["references"]]
                else:
                    return sev["xbrl-names"] if isinstance(sev["xbrl-names"], list) else [sev["xbrl-names"]]
        return []

    def _getAxesFromSev(self, sev):
        if sev.get("validation") == "noDups":
            return sev.get("axis", "").split("-")
        if sev.get("axis", sev.get("references-axes")) != sev.get("references-axes", sev.get("axis")):
            # assuming mismatching references-axes would indicate the main axis is being checked.
            return sev.get("axis", "").split("-")
        messageID, messageTemplate = self._getMessageFromSev(sev)
        if messageID:
            if "{tag}" in messageID:
                return sev.get("axis", "").split("-")
            elif "{otherTag}" in messageID:
                return sev.get("references-axes", sev.get("axis", "")).split("-")
            elif sev.get("msgCoda"):
                if "{tag}" in sev.get("msgCoda"):
                    return sev.get("axis", "").split("-")
                elif "{otherTag}" in sev.get("msgCoda"):
                    return sev.get("references-axes", sev.get("axis", "")).split("-")
                else:
                    return sev.get("axis", "").split("-")
        return []

    def _getRuleReferenceText(self, name):
        for tableFlags in self.RULES.values():
            for ruleFlag, ruleText in tableFlags.items():
                if ruleFlag == name:
                    return ruleText

    def _getRulesFromSev(self, sev, table):
        rules = []
        for rule, ruleText in self.RULES[table].items():
            ffdRule = f"ffd:{rule}"
            if ffdRule in sev.get("xbrl-names", []):
                rules.append(ruleText)
            elif ffdRule in sev.get("where", []):
                if True in sev["where"][ffdRule] and not "!not!" in sev["where"][ffdRule]:
                    rules.append(ruleText)
            elif ffdRule in sev.get("references-where", []):
                if True in sev["references-where"][ffdRule] and not "!not!" in sev["references-where"][ffdRule]:
                    rules.append(ruleText)
        if not rules:
            for rule, ruleText in self.RULES[table].items():
                ffdRule = f"ffd:{rule}"
                ruleWhere = sev.get("where", {}).get(ffdRule, []) + sev.get("references-where", {}).get(ffdRule, [])
                if (not "absent" in ruleWhere and not False in ruleWhere) and not (True in ruleWhere and "!not!" in ruleWhere):
                    rules.append(ruleText)
        return rules

    def _getXbrlNameExNS(self, xbrlName):
        return xbrlName.split(":")[-1]

    def _getRuleReference(self, xbrlName, sev):
        xbrlNameExNS = self._getXbrlNameExNS(xbrlName)
        axes = self._getAxesFromSev(sev)
        if sev.get("efm") == "ft.usrSty42":
            return ["Rule Flag Issue"]
        if xbrlName.endswith("Flg") and "Rule" in xbrlName or sev.get("validation", "").startswith("exist-in-axis"):
            if "usrSty" in sev.get("efm"):
                usrStyNumber = sev.get("efm").split(".")[-1].replace("usrSty", "")
                return [f"User Story {usrStyNumber}"]
            return ["Rule Flag Issue"]
        if sev.get("validation") == "noDups":
            if "of" in axes:
                return self._getRulesFromSev(sev, "Offering")
            if "o" in axes:
                return self._getRulesFromSev(sev, "Offset")
            if "cp" in axes:
                return self._getRulesFromSev(sev, "Combined Prospectus")
            if "s424i" in axes:
                return ["Securities, 424I"]
        if xbrlNameExNS in ftSubTbl or xbrlNameExNS in deiSubTblCodes:
            return ["Submission Table"]
        if xbrlNameExNS in ftSumTbl:
            return ["Fees Summary"]
        if (xbrlNameExNS in ftOfferingTbl or "header:" in xbrlName) and "of" in axes:
            return self._getRulesFromSev(sev, "Offering")
        if xbrlNameExNS in ftOffsetTbl and "o" in axes:
            return self._getRulesFromSev(sev, "Offset")
        if xbrlNameExNS in ftCmbPrsTbl and "cp" in axes:
            return self._getRulesFromSev(sev, "Combined Prospectus")
        if xbrlNameExNS in ft424iTbl and "s424i" in axes:
            return ["Securities, 424I"]
        return []

    def _getDataElement(self, xbrlName, sev):
        if sev.get("validation") == "noDups":
            return xbrlName
        if xbrlName.startswith("header:"):
            return xbrlName
        if ", " in xbrlName:
            labels = []
            for name in xbrlName.split(", "):
                xbrlNameExNS = self._getXbrlNameExNS(name)
                concepts = self.cntlr.modelManager.modelXbrl.nameConcepts.get(xbrlNameExNS, "")
                if concepts:
                    labels.append(concepts[0].label(XbrlConst.terseLabel))
            return ", ".join(labels)
        xbrlNameExNS = self._getXbrlNameExNS(xbrlName)
        concepts = self.cntlr.modelManager.modelXbrl.nameConcepts.get(xbrlNameExNS, "")
        if concepts:
            return concepts[0].label(XbrlConst.terseLabel)
        return concepts

    def _getLineType(self, xbrlName, sev, flagName):
        # returns tuple with 2 strings.
        # If the flagName should only be true ("X", "")
        # If the flagName can be true and false ("X", "X")
        # If the flagName not specified in whereClause ("X", "X")
        if flagName in sev.get("where", {}):
            whereValues = sev["where"][flagName]
            if True in whereValues and False in whereValues:
                return ("X", "X")
            if True in whereValues:
                return ("X", "")
        return ("X", "X")

    def _getValidationLineType(self, xbrlName, sev):
        xbrlNameExNS = self._getXbrlNameExNS(xbrlName)
        lineTypes = {
            "Fees to be Paid": "",
            "Fees Previously Paid": "",
            "Offset Claim": "",
            "Offset Source": ""
        }
        if xbrlNameExNS in ftOfferingTbl:
            lineTypes["Fees Previously Paid"], lineTypes["Fees to be Paid"] = self._getLineType(xbrlName, sev, "ffd:PrevslyPdFlg")
        if xbrlNameExNS in ftOffsetTbl:
            lineTypes["Offset Claim"], lineTypes["Offset Source"] = self._getLineType(xbrlName, sev, "ffd:OffsetClmdInd")
        return lineTypes

    def _getTableNamesFromSev(self, sev):
        tableNames = []
        for axis in self._getAxesFromSev(sev):
            if axis == "of":
                tableNames.append("Offering")
            elif axis == "o":
                tableNames.append("Offset")
            elif axis == "cp":
                tableNames.append("Combined Prospectus")
            elif axis == "s424i":
                tableNames.append("Securities, 424I")
        return tableNames

    def _getMessageNumericID(self, xbrlName, sev, ruleReference):
        msgNumId = ftStart
        subTypes = self._getSubTypes(sev)
        xbrlNameExNS = self._getXbrlNameExNS(xbrlName)
        subTypesList = [st.strip() for st in subTypes.split(",")] if subTypes else []

        if not "ft" in sev.get("efm"):
            msgNumId = efmStart
            msgNumId += sum(int(n) * 10**(i*2) for i,n in enumerate(reversed(sev.get("efm", "").split(".")), start=1))
            return [(msgNumId, subTypes)]

        validationName = sev.get("efm").split(".")[-1]
        validationNumber = ftValidations.get(validationName, 0)
        numericIDs = []

        if sev.get("validation") == "fdepflag-any" and len(self._getAxesFromSev(sev)) > 1:
            return [(msgNumId + validationNumber, subTypes)]

        if xbrlNameExNS in deiSubTblCodes:
            msgNumId = deiSubTblCodes[xbrlNameExNS]
            return [(msgNumId + validationNumber, subTypes)]
        elif xbrlNameExNS in ftSubTbl:
            msgNumId += ftTableStartCode[0][0]
            msgNumId += (ftSubTbl.index(xbrlNameExNS) + 1) * 100
            msgNumId += validationNumber
            if "424I" in subTypesList:
                numericIDs.append((msgNumId + 20000, "424I"))
                if len(subTypesList) == 1:
                    return numericIDs
            numericIDs.append((msgNumId + 10000, ", ".join([st for st in subTypesList if st != "424I"])))
            return numericIDs
        elif xbrlNameExNS in ftSumTbl:
            msgNumId += ftTableStartCode[0][0] + 1000000
            msgNumId += (ftSumTbl.index(xbrlNameExNS) + 1) * 100
            msgNumId += validationNumber
            scSubTypes = []
            nonSc424SubTypes = []
            posSubTypes = []
            SubTypes424B = []
            for st in subTypesList:
                if st.startswith("SC") or st.startswith("PRER") or st.startswith("PREM"):
                    scSubTypes.append(st)
                if not (st.startswith("SC") or st.startswith("PRER") or st.startswith("PREM")) and not st == "424I":
                    nonSc424SubTypes.append(st)
                if st.startswith("POS"):
                    posSubTypes.append(st)
                if st.startswith("424B"):
                    SubTypes424B.append(st)

            if "424I" in subTypesList:
                numericIDs.append((msgNumId + 10000, "424I"))
            if scSubTypes:
                numericIDs.append((msgNumId + 20000, ", ".join(scSubTypes)))
            if ftSumTbl.index(xbrlNameExNS) + 1 <= 8 and nonSc424SubTypes:
                numericIDs.append((msgNumId + 30000, ", ".join(nonSc424SubTypes)))
            else:
                if posSubTypes:
                    numericIDs.append((msgNumId + 40000, ", ".join(posSubTypes)))
                if SubTypes424B:
                    numericIDs.append((msgNumId + 50000, ", ".join(SubTypes424B)))
                if nonSc424SubTypes:
                    numericIDs.append((msgNumId, ", ".join(nonSc424SubTypes)))
            return numericIDs
        ruleCode = 0
        if not (xbrlNameExNS.endswith("Flg") and "Rule" in xbrlNameExNS):
            for (ruleNumber, ruleText) in ftRuleCode:
                if ruleText in ruleReference:
                    ruleCode = ruleNumber
                    break
        tableNames = self._getTableNamesFromSev(sev)
        for (code, tblName, tblConcepts) in ftTableStartCode:
            if tblConcepts == None:
                continue
            if sev.get("validation") == "noDups" and tblName in xbrlNameExNS:
                numericIDs.append((msgNumId + code + ruleCode + validationNumber, subTypes))
                break
            if (xbrlNameExNS in tblConcepts or "header:" in xbrlName) and tblName in tableNames:
                conceptCode = 0 if "header:" in xbrlName else (tblConcepts.index(xbrlNameExNS) + 1) * 100
                numericIDs.append((msgNumId + code + ruleCode + conceptCode + validationNumber, subTypes))
            elif xbrlNameExNS in self.RULES.get(tblName, {}):
                numericIDs.append((msgNumId + code + validationNumber, subTypes))
        return numericIDs

    def _getValidationSeverity(self, sev):
        severityMapping = {
            "warning": "W",
            "error": "E",
            "info": "I"
        }
        if sev.get("severity"):
            return severityMapping.get(sev["severity"])
        if sev.get("validation"):
            return severityMapping.get(self.validationsJson["validations"].get(sev["validation"], {}).get("severity"))

    def _processMessageArgs(self, messageTemplate, sev):
        messageArgs = self.validationsJson["validations"][sev["validation"]]
        for messageArg, value in messageArgs.items():
            if sev.get(messageArg) and messageArg != "value":
                messageTemplate = messageTemplate.replace(f"{{{messageArg}}}", sev.get(messageArg))
            else:
                messageTemplate = messageTemplate.replace(f"{{{messageArg}}}", value)
        if "{severityVerb}" in messageTemplate:
            severity = sev.get("severity") or messageArgs.get("severity")
            severityVerb = {"warning": "should", "error": "must", "info": "may"}.get(severity, "{severityVerb}")
            messageTemplate = messageTemplate.replace("{severityVerb}", severityVerb)
        if "{msgCoda}" in messageTemplate and sev.get("msgCoda"):
            messageTemplate = messageTemplate.replace("{msgCoda}", sev["msgCoda"])
        if "{comparison}" in messageTemplate:
            comparison = sev.get("comparison")
            comparisonText = sev.get("comparisonText", messageArgs.get("comparisonText", comparison)).format(comparison=comparison)
            messageTemplate = messageTemplate.replace("{comparison}", comparisonText)
        if "{expectedValue}" in messageTemplate:
            def getExpectValue(key):
                expectedValue = [json.dumps(value) for value in sev[key] if value != "!not!" ]
                expectedValue = "one of " + ", ".join(expectedValue) if len(expectedValue) > 1 else expectedValue[0]
                return expectedValue
            if sev.get("value"):
                expectedValue = getExpectValue("value")
                messageTemplate = messageTemplate.replace("{expectedValue}", expectedValue)
            elif sev.get("validation") == "fdepflag-ref-value":
                expectedValue = getExpectValue("reference-value")
                messageTemplate = messageTemplate.replace("{expectedValue}", expectedValue)
            elif sev.get("value-date-range") or sev.get("value-numeric-range"):
                rangeKey = "value-date-range" if sev.get("value-date-range") else "value-numeric-range"
                expectedValue = f"from {sev[rangeKey][0]} to {sev[rangeKey][1]}"
                messageTemplate = messageTemplate.replace("{expectedValue}", expectedValue)
        return messageTemplate

    def _getValidationMessage(self, xbrlName, sev):
        messageID, messageTemplate = self._getMessageFromSev(sev)
        if messageID:
            xbrlNameExNS = self._getXbrlNameExNS(xbrlName)
            if sev.get("msgCoda"):
                messageTemplate = messageTemplate.replace("{msgCoda}", sev["msgCoda"])
            else:
                messageTemplate = messageTemplate.replace("{msgCoda}", "")

            if "{tag}" in messageID:
                messageID = messageID.replace("{tag}", xbrlNameExNS)
                messageTemplate = messageTemplate.replace("{tag}", xbrlName)
            elif "{tags}" in messageID:
                names = sev.get("xbrl-names", []) if isinstance(sev.get("xbrl-names", []), list) else [sev.get("xbrl-names")]
                xbrlNamesExNS = [self._getXbrlNameExNS(xbrlNm) for xbrlNm in names]
                messageID = messageID.replace("{tags}", "-".join(xbrlNamesExNS))
                messageTemplate = messageTemplate.replace("{tags}", ", ".join(names))
            elif "{otherTag}" in messageID:
                if sev.get("validation") == "fdepflag-any":
                    names = sev.get("references", []) if isinstance(sev.get("references", []), list) else [sev.get("references")]
                    messageID = messageID.replace("{otherTag}", "-".join([self._getXbrlNameExNS(xbrlNm) for xbrlNm in names]))
                    messageTemplate = messageTemplate.replace("{otherTag}", ", ".join([xbrlNm for xbrlNm in sev.get("references", [])]))
                else:
                    messageID = messageID.replace("{otherTag}", xbrlNameExNS)
                    messageTemplate = messageTemplate.replace("{otherTag}", xbrlName)
            elif sev.get("msgCoda"):
                if "{tag}" in sev.get("msgCoda"):
                    messageTemplate = messageTemplate.replace("{tag}", xbrlName)
                elif "{tags}" in sev.get("msgCoda"):
                    names = sev.get("xbrl-names", []) if isinstance(sev.get("xbrl-names", []), list) else [sev.get("xbrl-names")]
                    messageTemplate = messageTemplate.replace("{tags}", ", ".join(names))
                elif "{otherTag}" in sev.get("msgCoda"):
                    messageTemplate = messageTemplate.replace("{otherTag}", xbrlName)
            if "{otherTag}" in messageTemplate:
                names = sev.get("references", []) if isinstance(sev.get("references", []), list) else [sev.get("references")]
                messageTemplate = messageTemplate.replace("{otherTag}", ", ".join(names))
            messageTemplate = self._processMessageArgs(messageTemplate, sev)
            if "{item}" in messageTemplate:
                names = sev.get("references", []) if isinstance(sev.get("references", []), list) else [sev.get("references")]
                messageTemplate = messageTemplate.replace("{item}", ", ".join(names))
            if "{otherLabel}" in messageTemplate:
                names = sev.get("references", []) if isinstance(sev.get("references", []), list) else [sev.get("references")]
                label = self._getDataElement(", ".join(names), sev)
                messageTemplate = messageTemplate.replace("{otherLabel}", label)
        return f"{messageID}: {messageTemplate}"

    def _getSubTypes(self, sev):
        if sev.get("sub-types", "n/a") in ["n/a", ["n/a"]]:
            subTypes = []
            for subType in self.submissionTypeClasses["all-Ixbrl"]:
                subTypes.append(subType)
            return ", ".join(subTypes)
        else:
            subTypes = []
            if "!not!" in sev["sub-types"]:
                subTypesToExclude = set()

                for subType in sev["sub-types"]:
                    if subType.startswith("@"):
                        for st in self.submissionTypeClasses[subType[1:]]:
                            subTypesToExclude.add(st)
                    else:
                        subTypesToExclude.add(subType)

                for subType in self.submissionTypeClasses["all-Ixbrl"]:
                    if not subType in subTypesToExclude:
                        subTypes.append(subType)
            else:
                for subType in sev["sub-types"]:
                    if subType.startswith("@"):
                        for st in self.submissionTypeClasses[subType[1:]]:
                            subTypes.append(st)
                    else:
                        subTypes.append(subType)
            return ", ".join(subTypes)
        return ""



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional

from arelle.utils.PluginHooks import ValidationHook
from arelle.utils.validate.Validation import Validation

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

ValidationFunction: TypeAlias = Callable[..., Optional[Iterable[Validation]]]


_VALIDATION_RULE_ATTRIBUTES_KEY = "_ARELLE_VALIDATION_ATTRIBUTES"


def validation(
    hook: ValidationHook,
    disclosureSystems: str | list[str] | None = None,
    excludeDisclosureSystems: str | list[str] | None = None,
) -> Callable[[ValidationFunction], ValidationFunction]:
    """
    Decorator for registering plugin validations. At most one of disclosureSystems and
    excludeDisclosureSystems parameters may be provided. If neither is provided the rule
    applies to all disclosure systems.

    All validation plugin functions should begin with the prefix "rule". For instance `def rule05(...):`.
    There is a test that checks all validation plugin functions (in the Arelle repo) beginning with the
    "rule" prefix to verify that the decorator has been applied.

    :param hook: The plugin validation hook the function should run with.
    :param disclosureSystems: The disclosure systems the validation should run with.
    :param excludeDisclosureSystems: The disclosure systems to exclude from running the rule with.
    :return: the registered validation function.
    """
    parsedIncluded = _wrapStrWithList(disclosureSystems)
    parsedExcluded = _wrapStrWithList(excludeDisclosureSystems) or []
    attributes = ValidationAttributes(hook, parsedIncluded, parsedExcluded)

    def decorator(f: ValidationFunction) -> ValidationFunction:
        if not callable(f):
            raise ValueError("@validation decorator must be used only to decorate functions.")

        _setValidationAttributes(f, attributes)
        return f

    return decorator


def _wrapStrWithList(items: list[str] | str | None) -> list[str] | None:
    return [items] if isinstance(items, str) else items


@dataclass(frozen=True)
class ValidationAttributes:
    hook: ValidationHook
    disclosureSystems: list[str] | None
    excludeDisclosureSystems: list[str]

    def __post_init__(self) -> None:
        if self.disclosureSystems is not None:
            if len(self.disclosureSystems) == 0:
                raise ValueError("disclosureSystems is an empty list. You can provide a list of disclosure systems, use None to match all, or use None and excludeDisclosureSystems to select disclosure systems.")
            if len(self.excludeDisclosureSystems) > 0:
                raise ValueError("disclosureSystems or excludeDisclosureSystems may be used, but not both.")


def getValidationAttributes(func: Callable[..., Any]) -> ValidationAttributes | None:
    return getattr(func, _VALIDATION_RULE_ATTRIBUTES_KEY, None)


def _setValidationAttributes(func: Callable[..., Any], attributes: ValidationAttributes) -> None:
    setattr(func, _VALIDATION_RULE_ATTRIBUTES_KEY, attributes)



from pathlib import PurePath, Path
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig, ConformanceSuiteAssetConfig, AssetSource

CONFORMANCE_SUITE_ZIP_NAME = 'efm-70-240701.zip'

config = ConformanceSuiteConfig(
    additional_plugins_by_prefix=[(f'conf/{t}', frozenset({'EdgarRenderer'})) for t in [
        '612-presentation-syntax/612-09-presented-units-order',
        '626-rendering-syntax',
    ]],
    args=[
        '--disclosureSystem', 'efm-pragmatic',
        '--formula', 'run',
    ],
    assets=[
        ConformanceSuiteAssetConfig.conformance_suite(
            Path(CONFORMANCE_SUITE_ZIP_NAME),
            entry_point=Path('conf/testcases.xml'),
            public_download_url=f'https://www.sec.gov/files/edgar/{CONFORMANCE_SUITE_ZIP_NAME}',
            source=AssetSource.S3_PUBLIC,
        )
    ],
    cache_version_id='C5EhkW8aZ5CzXQ6Lvj8K8ezEYGYGoHUK',
    expected_failure_ids=frozenset([
        # Duplicated schema error. Extensible Enumerations 2.0 schema is accessed over http and https.
        'conf/622-only-supported-locations/622-01-all-supported-locations/622-01-all-supported-locations-testcase.xml:_010gd'
    ]),
    info_url='https://www.sec.gov/structureddata/osdinteractivedatatestsuite',
    name=PurePath(__file__).stem,
    plugins=frozenset({'validate/EFM', 'inlineXbrlDocumentSet'}),
    shards=40,
    test_case_result_options='match-any',
)



from __future__ import annotations

import itertools
import json
import urllib.request
import zipfile
from pathlib import Path
from shutil import rmtree

from arelle.RuntimeOptions import RuntimeOptions
from arelle.api.Session import Session
from arelle.plugin.validate import FERC
from tests.integration_tests.integration_test_util import get_s3_uri
from tests.integration_tests.scripts.script_util import parse_args, prepare_logfile


this_file = Path(__file__)
args = parse_args(
    this_file.stem,
    "Confirm taxonomy service API usage works as expected.",
    arelle=False,
)
arelle_offline = args.offline
working_directory = Path(args.working_directory)
test_directory = Path(args.test_directory)
cache_directory = test_directory / 'cache'
arelle_log_file = prepare_logfile(test_directory, this_file, ext='json')
entry_point_url = 'https://eCollection.ferc.gov/taxonomy/form60/2024-04-01/form/form60/form-60_2024-04-01.xsd'
remove_file = cache_directory / 'http' / 'www.xbrl.org' / '2005' / 'xbrldt-2005.xsd'
cache_zip_url = get_s3_uri(
    'ci/caches/scripts/python_api_taxonomy_service.zip',
    version_id='rSYFEO6UrUHEHl4zaRnFiZZKgIcDFKkY'
)
cache_zip_path = test_directory / 'cache.zip'

print(f"Downloading cache with missing labels file: {cache_zip_path}")
urllib.request.urlretrieve(cache_zip_url, cache_zip_path)
with zipfile.ZipFile(cache_zip_path, "r") as zip_ref:
    zip_ref.extractall(cache_directory)

print(f"Loading model: {entry_point_url}")
options = RuntimeOptions(
    cacheDirectory=str(cache_directory),
    disclosureSystemName='FERC',
    entrypointFile=entry_point_url,
    internetConnectivity='offline',
    logFile='logToStructuredMessage',
    logLevel='DEBUG',
    logRefObjectProperties=True,
    plugins=FERC.__file__,
    utrValidate=True,
    validate=True,
)
with Session() as session:
    session.run(options)
    log_messages = session._cntlr.logHandler.messages
    with open(arelle_log_file, 'w') as f:
        json.dump(log_messages, f, ensure_ascii=False, indent=4)

print(f"Validating structured logs have errors for missing labels...")
message_code_groups = itertools.groupby(log_messages, key=lambda m: m['messageCode'])
expected_code_counts = {
    'info': 1,
    'IOerror': 1,
    'xbrl.5.2.4.2.1:preferredLabelMissing': 1910,
    'info:profileActivity': 1
}
for code, messages in message_code_groups:
    count = len(list(messages))
    assert expected_code_counts.get(code, []) == count, \
        f"Unexpected message count for {code}: {count}"

print("Cleaning up")
try:
    rmtree(working_directory / 'python_api_taxonomy_service' / 'cache')
    cache_zip_path.unlink()
except PermissionError as exc:
    print(f"Failed to cleanup test files: {exc}")



"""
See COPYRIGHT.md for copyright information.

Filer Guidelines:
- [Technical Note](https://www.revenue.ie/en/online-services/support/documents/ixbrl/ixbrl-technical-note.pdf)
- [Error Messages](https://www.revenue.ie/en/online-services/support/documents/ixbrl/error-messages.pdf)
- [Style Guide](https://www.revenue.ie/en/online-services/support/documents/ixbrl/ixbrl-style-guide.pdf)
"""
from __future__ import annotations

from arelle.Version import authorLabel, copyrightLabel
from pathlib import Path
from typing import Any
from .ValidationPluginExtension import ValidationPluginExtension
from .rules import ros

PLUGIN_NAME = "Validate ROS"
DISCLOSURE_SYSTEM_VALIDATION_TYPE = "ROS"

validationPlugin = ValidationPluginExtension(
    name=PLUGIN_NAME,
    disclosureSystemConfigUrl=Path(__file__).parent / "resources" / "config.xml",
    validationTypes=[DISCLOSURE_SYSTEM_VALIDATION_TYPE],
    validationRuleModules=[ros],
)


def disclosureSystemTypes(*args: Any, **kwargs: Any) -> tuple[tuple[str, str], ...]:
    return validationPlugin.disclosureSystemTypes


def disclosureSystemConfigURL(*args: Any, **kwargs: Any) -> str:
    return validationPlugin.disclosureSystemConfigURL


def validateXbrlStart(val, parameters=None, *args, **kwargs):
    val.validateROSplugin = val.validateDisclosureSystem and getattr(val.disclosureSystem, "ROSplugin", False)
    if not (val.validateROSplugin):
        return


def validateXbrlFinally(*args: Any, **kwargs: Any) -> None:
    return validationPlugin.validateXbrlFinally(*args, **kwargs)


__pluginInfo__ = {
    'name': PLUGIN_NAME,
    'version': '1.0',
    'description': '''ROS (Ireland) Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    'import': ('inlineXbrlDocumentSet', ), # import dependent modules
    # classes of mount points (required)
    'DisclosureSystem.Types': disclosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
    'Validate.XBRL.Start': validateXbrlStart,
    'Validate.XBRL.Finally': validateXbrlFinally,
}



from __future__ import annotations
from typing import Any


def test_conformance_suite(conformance_suite_results: dict[str, Any]) -> None:
    """
    See conftest.py for context around the parameterization of conformance suite results.
    It is critical that this file is not imported or referenced by other modules to ensure that it is not evaluated
    before pytest can evaluate conformance suite results via the pytest_configure hook.
    """
    assert conformance_suite_results.get('status') == 'pass', \
        'Expected these validation suffixes: {}, but received these validations: {}'.format(
            conformance_suite_results.get('expected'), conformance_suite_results.get('actual')
        )



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import logging
import sys
from typing import TextIO


class LogToPrintHandler(logging.Handler):
    """
    .. class:: LogToPrintHandler()

    A log handler that emits log entries to standard out as they are logged.

    CAUTION: Output is utf-8 encoded, which is fine for saving to files, but may not display correctly in terminal windows.

    :param logOutput: 'logToStdErr' to cause log print to stderr instead of stdout
    :type logOutput: str
    """
    logFile: str | TextIO | None

    def __init__(self, logOutput: str) -> None:
        super(LogToPrintHandler, self).__init__()
        if logOutput == "logToStdErr":
            self.logFile = sys.stderr
        else:
            self.logFile = None

    def emit(self, logRecord: logging.LogRecord) -> None:
        file = sys.stderr if self.logFile else None
        logEntry = self.format(logRecord)
        try:
            print(logEntry, file=file)
        except UnicodeEncodeError:
            # extra parentheses in print to allow for 3-to-2 conversion
            print((logEntry
                   .encode(sys.stdout.encoding, 'backslashreplace')
                   .decode(sys.stdout.encoding, 'strict')),
                  file=file)



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Any

from arelle.ModelDocument import LoadingException, ModelDocument
from arelle.ModelXbrl import ModelXbrl
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self, validateXbrl: ValidateXbrl) -> PluginValidationDataExtension:
        return PluginValidationDataExtension(self.name)

    def modelDocumentPullLoader(
        self,
        modelXbrl: ModelXbrl,
        normalizedUri: str,
        filepath: str,
        isEntry: bool,
        namespace: str | None,
        *args: Any,
        **kwargs: Any,
    ) -> ModelDocument | LoadingException | None:
        if self.disclosureSystemFromPluginSelected(modelXbrl):
            return LoadingException(_("XYZ validation plugin is a template for new validation plugins and shouldn't be used directly."))
        return None

    def modelXbrlLoadComplete(
        self,
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if self.disclosureSystemFromPluginSelected(modelXbrl):
            if modelXbrl.modelDocument is None:
                modelXbrl.error(
                    codes="XYZ.01.01",
                    msg=_("An XBRL Report Package is required but could not be loaded"),
                    modelObject=modelXbrl,
                )
        return None



"""
This module represents the time stamp when Arelle was last built

See COPYRIGHT.md for copyright information.

"""
from __future__ import annotations


def getBuildVersion() -> str | None:
    try:
        import arelle._version
        return arelle._version.version
    except ModuleNotFoundError:
        return None


def getGitHash() -> str | None:
    import subprocess
    try:
        return subprocess.run(['git', 'rev-parse', 'HEAD'], capture_output=True, check=True, text=True).stdout.strip()
    except (FileNotFoundError, subprocess.SubprocessError):
        return None


def getDefaultVersion() -> str:
    return '0.0.0'


def getVersion() -> str:
    for version_fetcher in [getBuildVersion, getGitHash, getDefaultVersion]:
        fetched_version = version_fetcher()
        if fetched_version is not None:
            return fetched_version
    raise ValueError('Version not set')


__version__ = getVersion()
version = __version__
authorLabel = 'Workiva, Inc.'
copyrightLabel = '(c) Copyright 2011-present Workiva, Inc., All rights reserved.'
